#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy
import datetime
from unittest import mock
import uuid

from keystoneauth1 import access
from keystoneauth1 import fixture
from openstack.identity.v3 import _proxy
from osc_lib.cli import format_columns

from openstackclient.tests.unit import fakes
from openstackclient.tests.unit import utils

base_url = 'http://identity:5000/v3/'

domain_id = 'd1'
domain_name = 'oftheking'
domain_description = 'domain description'

DOMAIN = {
    'id': domain_id,
    'name': domain_name,
    'description': domain_description,
    'enabled': True,
    'tags': [],
    'links': base_url + 'domains/' + domain_id,
}

group_id = 'gr-010'
group_name = 'spencer davis'

GROUP = {
    'id': group_id,
    'name': group_name,
    'links': base_url + 'groups/' + group_id,
}

mapping_id = 'test_mapping'
mapping_rules_file_path = '/tmp/path/to/file'
# Copied from
# (https://github.com/openstack/keystone/blob\
# master/keystone/tests/mapping_fixtures.py
EMPLOYEE_GROUP_ID = "0cd5e9"
DEVELOPER_GROUP_ID = "xyz"
MAPPING_RULES = [
    {
        "local": [{"group": {"id": EMPLOYEE_GROUP_ID}}],
        "remote": [
            {"type": "orgPersonType", "not_any_of": ["Contractor", "Guest"]}
        ],
    }
]

MAPPING_RULES_2 = [
    {
        "local": [{"group": {"id": DEVELOPER_GROUP_ID}}],
        "remote": [{"type": "orgPersonType", "any_one_of": ["Contractor"]}],
    }
]


MAPPING_RESPONSE = {"id": mapping_id, "rules": MAPPING_RULES}

MAPPING_RESPONSE_2 = {"id": mapping_id, "rules": MAPPING_RULES_2}

mfa_opt1 = 'password,totp'
mfa_opt2 = 'password'

project_id = '8-9-64'
project_name = 'beatles'
project_description = 'Fab Four'

PROJECT = {
    'id': project_id,
    'name': project_name,
    'description': project_description,
    'enabled': True,
    'domain_id': domain_id,
    'tags': [],
    'links': base_url + 'projects/' + project_id,
}

PROJECT_2 = {
    'id': project_id + '-2222',
    'name': project_name + ' reprise',
    'description': project_description + 'plus four more',
    'enabled': True,
    'domain_id': domain_id,
    'tags': [],
    'links': base_url + 'projects/' + project_id,
}

region_id = 'region_one'
region_parent_region_id = 'region_two'
region_description = 'region one'

REGION = {
    'id': region_id,
    'description': region_description,
    'parent_region_id': region_parent_region_id,
    'links': base_url + 'regions/' + region_id,
}

PROJECT_WITH_PARENT = {
    'id': project_id + '-with-parent',
    'name': project_name + ' and their parents',
    'description': project_description + ' plus another four',
    'enabled': True,
    'domain_id': domain_id,
    'parent_id': project_id,
    'tags': [],
    'links': base_url + 'projects/' + (project_id + '-with-parent'),
}

PROJECT_WITH_GRANDPARENT = {
    'id': project_id + '-with-grandparent',
    'name': project_name + ', granny and grandpa',
    'description': project_description + ' plus another eight?',
    'enabled': True,
    'domain_id': domain_id,
    'parent_id': PROJECT_WITH_PARENT['id'],
    'tags': [],
    'links': base_url + 'projects/' + (project_id + '-with-grandparent'),
}

parents = [{'project': PROJECT}]
grandparents = [{'project': PROJECT}, {'project': PROJECT_WITH_PARENT}]
ids_for_parents = [PROJECT['id']]
ids_for_parents_and_grandparents = [PROJECT['id'], PROJECT_WITH_PARENT['id']]

children = [{'project': PROJECT_WITH_GRANDPARENT}]
ids_for_children = [PROJECT_WITH_GRANDPARENT['id']]


role_id = 'r1'
role_name = 'roller'
role_description = 'role description'

ROLE = {
    'id': role_id,
    'name': role_name,
    'domain': None,
    'links': base_url + 'roles/' + role_id,
}

ROLE_2 = {
    'id': 'r2',
    'name': 'Rolls Royce',
    'domain': domain_id,
    'links': base_url + 'roles/' + 'r2',
}

ROLES = [ROLE, ROLE_2]

service_id = 's-123'
service_name = 'Texaco'
service_type = 'gas'
service_description = 'oil brand'

SERVICE = {
    'id': service_id,
    'name': service_name,
    'type': service_type,
    'description': service_description,
    'enabled': True,
    'links': base_url + 'services/' + service_id,
}

SERVICE_WITHOUT_NAME = {
    'id': service_id,
    'type': service_type,
    'description': service_description,
    'enabled': True,
    'links': base_url + 'services/' + service_id,
}

endpoint_id = 'e-123'
endpoint_url = 'http://127.0.0.1:35357'
endpoint_region = 'RegionOne'
endpoint_interface = 'admin'

ENDPOINT = {
    'id': endpoint_id,
    'url': endpoint_url,
    'region': endpoint_region,
    'interface': endpoint_interface,
    'service_id': service_id,
    'enabled': True,
    'links': base_url + 'endpoints/' + endpoint_id,
}

endpoint_group_id = 'eg-123'
endpoint_group_description = 'eg 123 description'
endpoint_group_filters = {
    'service_id': service_id,
    'region_id': endpoint_region,
}
endpoint_group_filters_2 = {
    'region_id': endpoint_region,
}
endpoint_group_file_path = '/tmp/path/to/file'

ENDPOINT_GROUP = {
    'id': endpoint_group_id,
    'filters': endpoint_group_filters,
    'description': endpoint_group_description,
    'links': base_url + 'endpoint_groups/' + endpoint_group_id,
}

user_id = 'bbbbbbb-aaaa-aaaa-aaaa-bbbbbbbaaaa'
user_name = 'paul'
user_description = 'Sir Paul'
user_email = 'paul@applecorps.com'

USER = {
    'id': user_id,
    'name': user_name,
    'default_project_id': project_id,
    'email': user_email,
    'enabled': True,
    'domain_id': domain_id,
    'links': base_url + 'users/' + user_id,
}

trust_id = 't-456'
trust_expires = None
trust_impersonation = False
trust_roles = ({"id": role_id, "name": role_name},)

TRUST = {
    'expires_at': trust_expires,
    'id': trust_id,
    'impersonation': trust_impersonation,
    'links': base_url + 'trusts/' + trust_id,
    'project_id': project_id,
    'roles': trust_roles,
    'trustee_user_id': user_id,
    'trustor_user_id': user_id,
}

token_expires = '2016-09-05T18:04:52+0000'
token_id = 'tttttttt-tttt-tttt-tttt-tttttttttttt'

UNSCOPED_TOKEN = {
    'expires': token_expires,
    'id': token_id,
    'user_id': user_id,
}

TOKEN_WITH_PROJECT_ID = {
    'expires': token_expires,
    'id': token_id,
    'project_id': project_id,
    'user_id': user_id,
}

TOKEN_WITH_DOMAIN_ID = {
    'expires': token_expires,
    'id': token_id,
    'domain_id': domain_id,
    'user_id': user_id,
}

idp_id = 'test_idp'
idp_description = 'super exciting IdP description'
idp_remote_ids = ['entity1', 'entity2']
formatted_idp_remote_ids = format_columns.ListColumn(idp_remote_ids)

IDENTITY_PROVIDER = {
    'id': idp_id,
    'remote_ids': idp_remote_ids,
    'enabled': True,
    'description': idp_description,
    'domain_id': domain_id,
}

protocol_id = 'protocol'

mapping_id = 'test_mapping'
mapping_id_updated = 'prod_mapping'

sp_id = 'BETA'
sp_description = 'Service Provider to burst into'
service_provider_url = 'https://beta.example.com/Shibboleth.sso/POST/SAML'
sp_auth_url = (
    'https://beta.example.com/v3/OS-FEDERATION/identity_providers/'
    'idp/protocol/saml2/auth'
)

SERVICE_PROVIDER = {
    'id': sp_id,
    'enabled': True,
    'description': sp_description,
    'sp_url': service_provider_url,
    'auth_url': sp_auth_url,
}

PROTOCOL_ID_MAPPING = {'id': protocol_id, 'mapping': mapping_id}

PROTOCOL_OUTPUT = {
    'id': protocol_id,
    'mapping_id': mapping_id,
    'identity_provider': idp_id,
}

PROTOCOL_OUTPUT_UPDATED = {
    'id': protocol_id,
    'mapping_id': mapping_id_updated,
    'identity_provider': idp_id,
}

# Assignments

ASSIGNMENT_WITH_PROJECT_ID_AND_USER_ID = {
    'scope': {'project': {'id': project_id}},
    'user': {'id': user_id},
    'role': {'id': role_id},
}

ASSIGNMENT_WITH_PROJECT_ID_AND_USER_ID_INCLUDE_NAMES = {
    'scope': {
        'project': {
            'domain': {'id': domain_id, 'name': domain_name},
            'id': project_id,
            'name': project_name,
        }
    },
    'user': {
        'domain': {'id': domain_id, 'name': domain_name},
        'id': user_id,
        'name': user_name,
    },
    'role': {'id': role_id, 'name': role_name},
}

ASSIGNMENT_WITH_PROJECT_ID_AND_USER_ID_INHERITED = {
    'scope': {
        'project': {'id': project_id},
        'OS-INHERIT:inherited_to': 'projects',
    },
    'user': {'id': user_id},
    'role': {'id': role_id},
}

ASSIGNMENT_WITH_PROJECT_ID_AND_GROUP_ID = {
    'scope': {'project': {'id': project_id}},
    'group': {'id': group_id},
    'role': {'id': role_id},
}

ASSIGNMENT_WITH_DOMAIN_ID_AND_USER_ID = {
    'scope': {'domain': {'id': domain_id}},
    'user': {'id': user_id},
    'role': {'id': role_id},
}

ASSIGNMENT_WITH_DOMAIN_ROLE = {
    'scope': {'domain': {'id': domain_id}},
    'user': {'id': user_id},
    'role': {'id': ROLE_2['id']},
}

ASSIGNMENT_WITH_DOMAIN_ID_AND_USER_ID_INCLUDE_NAMES = {
    'scope': {'domain': {'id': domain_id, 'name': domain_name}},
    'user': {
        'domain': {'id': domain_id, 'name': domain_name},
        'id': user_id,
        'name': user_name,
    },
    'role': {'id': role_id, 'name': role_name},
}

ASSIGNMENT_WITH_DOMAIN_ID_AND_USER_ID_INHERITED = {
    'scope': {
        'domain': {'id': domain_id},
        'OS-INHERIT:inherited_to': 'projects',
    },
    'user': {'id': user_id},
    'role': {'id': role_id},
}

ASSIGNMENT_WITH_DOMAIN_ID_AND_GROUP_ID = {
    'scope': {'domain': {'id': domain_id}},
    'group': {'id': group_id},
    'role': {'id': role_id},
}

consumer_id = 'test consumer id'
consumer_description = 'someone we trust'
consumer_secret = 'test consumer secret'

OAUTH_CONSUMER = {
    'id': consumer_id,
    'secret': consumer_secret,
    'description': consumer_description,
}

access_token_id = 'test access token id'
access_token_secret = 'test access token secret'
access_token_expires = '2014-05-18T03:13:18.152071Z'

OAUTH_ACCESS_TOKEN = {
    'id': access_token_id,
    'expires': access_token_expires,
    'key': access_token_id,
    'secret': access_token_secret,
}

request_token_id = 'test request token id'
request_token_secret = 'test request token secret'
request_token_expires = '2014-05-17T11:10:51.511336Z'

OAUTH_REQUEST_TOKEN = {
    'id': request_token_id,
    'expires': request_token_expires,
    'key': request_token_id,
    'secret': request_token_secret,
}

oauth_verifier_pin = '6d74XaDS'
OAUTH_VERIFIER = {'oauth_verifier': oauth_verifier_pin}

app_cred_id = 'app-cred-id'
app_cred_name = 'testing_app_cred'
app_cred_role = ({"id": role_id, "name": role_name, "domain": None},)
app_cred_description = 'app credential for testing'
app_cred_expires = datetime.datetime(2022, 1, 1, 0, 0)
app_cred_expires_str = app_cred_expires.strftime('%Y-%m-%dT%H:%M:%S%z')
app_cred_secret = 'moresecuresecret'
app_cred_access_rules = (
    '[{"path": "/v2.1/servers", "method": "GET", "service": "compute"}]'
)
app_cred_access_rules_path = '/tmp/access_rules.json'
access_rule_id = 'access-rule-id'
access_rule_service = 'compute'
access_rule_path = '/v2.1/servers'
access_rule_method = 'GET'
APP_CRED_BASIC = {
    'id': app_cred_id,
    'name': app_cred_name,
    'project_id': project_id,
    'roles': app_cred_role,
    'description': None,
    'expires_at': None,
    'unrestricted': False,
    'secret': app_cred_secret,
    'access_rules': None,
}
APP_CRED_OPTIONS = {
    'id': app_cred_id,
    'name': app_cred_name,
    'project_id': project_id,
    'roles': app_cred_role,
    'description': app_cred_description,
    'expires_at': app_cred_expires_str,
    'unrestricted': False,
    'secret': app_cred_secret,
    'access_rules': None,
}
ACCESS_RULE = {
    'id': access_rule_id,
    'service': access_rule_service,
    'path': access_rule_path,
    'method': access_rule_method,
}
APP_CRED_ACCESS_RULES = {
    'id': app_cred_id,
    'name': app_cred_name,
    'project_id': project_id,
    'roles': app_cred_role,
    'description': None,
    'expires_at': None,
    'unrestricted': False,
    'secret': app_cred_secret,
    'access_rules': app_cred_access_rules,
}

registered_limit_id = 'registered-limit-id'
registered_limit_default_limit = 10
registered_limit_description = 'default limit of foobars'
registered_limit_resource_name = 'foobars'
REGISTERED_LIMIT = {
    'id': registered_limit_id,
    'default_limit': registered_limit_default_limit,
    'resource_name': registered_limit_resource_name,
    'service_id': service_id,
    'description': None,
    'region_id': None,
}
REGISTERED_LIMIT_OPTIONS = {
    'id': registered_limit_id,
    'default_limit': registered_limit_default_limit,
    'resource_name': registered_limit_resource_name,
    'service_id': service_id,
    'description': registered_limit_description,
    'region_id': region_id,
}

limit_id = 'limit-id'
limit_resource_limit = 15
limit_description = 'limit of foobars'
limit_resource_name = 'foobars'
LIMIT = {
    'id': limit_id,
    'project_id': project_id,
    'resource_limit': limit_resource_limit,
    'resource_name': limit_resource_name,
    'service_id': service_id,
    'description': None,
    'region_id': None,
}
LIMIT_OPTIONS = {
    'id': limit_id,
    'project_id': project_id,
    'resource_limit': limit_resource_limit,
    'resource_name': limit_resource_name,
    'service_id': service_id,
    'description': limit_description,
    'region_id': region_id,
}


def fake_auth_ref(fake_token, fake_service=None):
    """Create an auth_ref using keystoneauth's fixtures"""
    token_copy = copy.deepcopy(fake_token)
    token_id = token_copy.pop('id')
    token = fixture.V3Token(**token_copy)
    # An auth_ref is actually an access info object
    auth_ref = access.create(
        body=token,
        auth_token=token_id,
    )

    # Create a service catalog
    if fake_service:
        service = token.add_service(
            fake_service['type'],
            fake_service['name'],
        )
        # TODO(dtroyer): Add an 'id' element to KSA's _Service fixure
        service['id'] = fake_service['id']
        for e in fake_service['endpoints']:
            region = e.get('region_id') or e.get('region', '<none>')
            service.add_endpoint(
                e['interface'],
                e['url'],
                region=region,
            )

    return auth_ref


class FakeAuth:
    def __init__(self, auth_method_class=None):
        self._auth_method_class = auth_method_class

    def get_token(self, *args, **kwargs):
        return token_id


class FakeSession:
    def __init__(self, **kwargs):
        self.auth = FakeAuth()


class FakeIdentityv3Client:
    def __init__(self, **kwargs):
        self.domains = mock.Mock()
        self.domains.resource_class = fakes.FakeResource(None, {})
        self.credentials = mock.Mock()
        self.credentials.resource_class = fakes.FakeResource(None, {})
        self.endpoints = mock.Mock()
        self.endpoints.resource_class = fakes.FakeResource(None, {})
        self.endpoint_filter = mock.Mock()
        self.endpoint_filter.resource_class = fakes.FakeResource(None, {})
        self.endpoint_groups = mock.Mock()
        self.endpoint_groups.resource_class = fakes.FakeResource(None, {})
        self.groups = mock.Mock()
        self.groups.resource_class = fakes.FakeResource(None, {})
        self.oauth1 = mock.Mock()
        self.oauth1.resource_class = fakes.FakeResource(None, {})
        self.projects = mock.Mock()
        self.projects.resource_class = fakes.FakeResource(None, {})
        self.regions = mock.Mock()
        self.regions.resource_class = fakes.FakeResource(None, {})
        self.roles = mock.Mock()
        self.roles.resource_class = fakes.FakeResource(None, {})
        self.services = mock.Mock()
        self.services.resource_class = fakes.FakeResource(None, {})
        self.session = mock.Mock()
        self.session.auth.auth_ref.service_catalog.resource_class = (
            fakes.FakeResource(None, {})
        )
        self.tokens = mock.Mock()
        self.tokens.resource_class = fakes.FakeResource(None, {})
        self.trusts = mock.Mock()
        self.trusts.resource_class = fakes.FakeResource(None, {})
        self.users = mock.Mock()
        self.users.resource_class = fakes.FakeResource(None, {})
        self.role_assignments = mock.Mock()
        self.role_assignments.resource_class = fakes.FakeResource(None, {})
        self.auth_token = kwargs['token']
        self.management_url = kwargs['endpoint']
        self.auth = FakeAuth()
        self.auth.client = mock.Mock()
        self.auth.client.resource_class = fakes.FakeResource(None, {})
        self.application_credentials = mock.Mock()
        self.application_credentials.resource_class = fakes.FakeResource(
            None, {}
        )
        self.access_rules = mock.Mock()
        self.access_rules.resource_class = fakes.FakeResource(None, {})
        self.inference_rules = mock.Mock()
        self.inference_rules.resource_class = fakes.FakeResource(None, {})
        self.registered_limits = mock.Mock()
        self.registered_limits.resource_class = fakes.FakeResource(None, {})
        self.limits = mock.Mock()
        self.limits.resource_class = fakes.FakeResource(None, {})


class FakeFederationManager:
    def __init__(self, **kwargs):
        self.identity_providers = mock.Mock()
        self.identity_providers.resource_class = fakes.FakeResource(None, {})
        self.mappings = mock.Mock()
        self.mappings.resource_class = fakes.FakeResource(None, {})
        self.protocols = mock.Mock()
        self.protocols.resource_class = fakes.FakeResource(None, {})
        self.projects = mock.Mock()
        self.projects.resource_class = fakes.FakeResource(None, {})
        self.domains = mock.Mock()
        self.domains.resource_class = fakes.FakeResource(None, {})
        self.service_providers = mock.Mock()
        self.service_providers.resource_class = fakes.FakeResource(None, {})


class FakeFederatedClient(FakeIdentityv3Client):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.federation = FakeFederationManager()


class FakeOAuth1Client(FakeIdentityv3Client):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.access_tokens = mock.Mock()
        self.access_tokens.resource_class = fakes.FakeResource(None, {})
        self.consumers = mock.Mock()
        self.consumers.resource_class = fakes.FakeResource(None, {})
        self.request_tokens = mock.Mock()
        self.request_tokens.resource_class = fakes.FakeResource(None, {})


class FakeClientMixin:
    def setUp(self):
        super().setUp()

        self.app.client_manager.identity = FakeIdentityv3Client(
            endpoint=fakes.AUTH_URL,
            token=fakes.AUTH_TOKEN,
        )
        self.identity_client = self.app.client_manager.identity

        # TODO(stephenfin): Rename to 'identity_client' once all commands are
        # migrated to SDK
        self.app.client_manager.sdk_connection.identity = mock.Mock(
            _proxy.Proxy
        )
        self.identity_sdk_client = (
            self.app.client_manager.sdk_connection.identity
        )


class TestIdentityv3(
    FakeClientMixin,
    utils.TestCommand,
): ...


class FakeModel(dict):
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)


# We don't use FakeClientMixin since we want a different fake legacy client
class TestFederatedIdentity(utils.TestCommand):
    def setUp(self):
        super().setUp()

        self.app.client_manager.identity = FakeFederatedClient(
            endpoint=fakes.AUTH_URL, token=fakes.AUTH_TOKEN
        )
        self.identity_client = self.app.client_manager.identity

        # TODO(stephenfin): Rename to 'identity_client' once all commands are
        # migrated to SDK
        self.app.client_manager.sdk_connection.identity = mock.Mock(
            _proxy.Proxy
        )
        self.identity_sdk_client = (
            self.app.client_manager.sdk_connection.identity
        )


# We don't use FakeClientMixin since we want a different fake legacy client
class TestOAuth1(utils.TestCommand):
    def setUp(self):
        super().setUp()

        self.app.client_manager.identity = FakeOAuth1Client(
            endpoint=fakes.AUTH_URL, token=fakes.AUTH_TOKEN
        )
        self.identity_client = self.app.client_manager.identity

        # TODO(stephenfin): Rename to 'identity_client' once all commands are
        # migrated to SDK
        self.app.client_manager.sdk_connection.identity = mock.Mock(
            _proxy.Proxy
        )
        self.identity_sdk_client = (
            self.app.client_manager.sdk_connection.identity
        )


class FakeProject:
    """Fake one or more project."""

    @staticmethod
    def create_one_project(attrs=None):
        """Create a fake project.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, name, and so on
        """

        attrs = attrs or {}

        # set default attributes.
        project_info = {
            'id': 'project-id-' + uuid.uuid4().hex,
            'name': 'project-name-' + uuid.uuid4().hex,
            'description': 'project-description-' + uuid.uuid4().hex,
            'enabled': True,
            'is_domain': False,
            'domain_id': 'domain-id-' + uuid.uuid4().hex,
            'parent_id': 'parent-id-' + uuid.uuid4().hex,
            'tags': [],
            'links': 'links-' + uuid.uuid4().hex,
        }
        project_info.update(attrs)

        project = fakes.FakeResource(
            info=copy.deepcopy(project_info), loaded=True
        )
        return project

    @staticmethod
    def create_projects(attrs=None, count=2):
        """Create multiple fake projects.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of projects to fake
        :return:
            A list of FakeResource objects faking the projects
        """

        projects = []
        for i in range(0, count):
            projects.append(FakeProject.create_one_project(attrs))
        return projects


class FakeDomain:
    """Fake one or more domain."""

    @staticmethod
    def create_one_domain(attrs=None):
        """Create a fake domain.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, name, and so on
        """

        attrs = attrs or {}

        # set default attributes.
        domain_info = {
            'id': 'domain-id-' + uuid.uuid4().hex,
            'name': 'domain-name-' + uuid.uuid4().hex,
            'description': 'domain-description-' + uuid.uuid4().hex,
            'enabled': True,
            'tags': [],
            'links': 'links-' + uuid.uuid4().hex,
        }
        domain_info.update(attrs)

        domain = fakes.FakeResource(
            info=copy.deepcopy(domain_info), loaded=True
        )
        return domain


class FakeCredential:
    """Fake one or more credential."""

    @staticmethod
    def create_one_credential(attrs=None):
        """Create a fake credential.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, type, and so on
        """

        attrs = attrs or {}

        # set default attributes.
        credential_info = {
            'id': 'credential-id-' + uuid.uuid4().hex,
            'type': 'cert',
            'user_id': 'user-id-' + uuid.uuid4().hex,
            'blob': 'credential-data-' + uuid.uuid4().hex,
            'project_id': 'project-id-' + uuid.uuid4().hex,
            'links': 'links-' + uuid.uuid4().hex,
        }
        credential_info.update(attrs)

        credential = fakes.FakeResource(
            info=copy.deepcopy(credential_info), loaded=True
        )
        return credential

    @staticmethod
    def create_credentials(attrs=None, count=2):
        """Create multiple fake credentials.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of credentials to fake
        :return:
            A list of FakeResource objects faking the credentials
        """
        credentials = []
        for i in range(0, count):
            credential = FakeCredential.create_one_credential(attrs)
            credentials.append(credential)

        return credentials

    @staticmethod
    def get_credentials(credentials=None, count=2):
        """Get an iterable MagicMock object with a list of faked credentials.

        If credentials list is provided, then initialize the Mock object with
        the list. Otherwise create one.

        :param List credentials:
            A list of FakeResource objects faking credentials
        :param Integer count:
            The number of credentials to be faked
        :return:
            An iterable Mock object with side_effect set to a list of faked
            credentials
        """
        if credentials is None:
            credentials = FakeCredential.create_credentials(count)

        return mock.Mock(side_effect=credentials)


class FakeUser:
    """Fake one or more user."""

    @staticmethod
    def create_one_user(attrs=None):
        """Create a fake user.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, name, and so on
        """

        attrs = attrs or {}

        # set default attributes.
        user_info = {
            'id': 'user-id-' + uuid.uuid4().hex,
            'name': 'user-name-' + uuid.uuid4().hex,
            'default_project_id': 'project-' + uuid.uuid4().hex,
            'email': 'user-email-' + uuid.uuid4().hex,
            'enabled': True,
            'domain_id': 'domain-id-' + uuid.uuid4().hex,
            'links': 'links-' + uuid.uuid4().hex,
        }
        user_info.update(attrs)

        user = fakes.FakeResource(info=copy.deepcopy(user_info), loaded=True)
        return user

    @staticmethod
    def create_users(attrs=None, count=2):
        """Create multiple fake users.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of users to fake
        :return:
            A list of FakeResource objects faking the users
        """
        users = []
        for i in range(0, count):
            user = FakeUser.create_one_user(attrs)
            users.append(user)

        return users

    @staticmethod
    def get_users(users=None, count=2):
        """Get an iterable MagicMock object with a list of faked users.

        If users list is provided, then initialize the Mock object with
        the list. Otherwise create one.

        :param List users:
            A list of FakeResource objects faking users
        :param Integer count:
            The number of users to be faked
        :return
            An iterable Mock object with side_effect set to a list of faked
            users
        """
        if users is None:
            users = FakeUser.create_users(count)

        return mock.Mock(side_effect=users)


class FakeGroup:
    """Fake one or more group."""

    @staticmethod
    def create_one_group(attrs=None):
        """Create a fake group.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, name, and so on
        """

        attrs = attrs or {}

        # set default attributes.
        group_info = {
            'id': 'group-id-' + uuid.uuid4().hex,
            'name': 'group-name-' + uuid.uuid4().hex,
            'links': 'links-' + uuid.uuid4().hex,
            'domain_id': 'domain-id-' + uuid.uuid4().hex,
            'description': 'group-description-' + uuid.uuid4().hex,
        }
        group_info.update(attrs)

        group = fakes.FakeResource(info=copy.deepcopy(group_info), loaded=True)
        return group

    @staticmethod
    def create_groups(attrs=None, count=2):
        """Create multiple fake groups.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of groups to fake
        :return:
            A list of FakeResource objects faking the groups
        """
        groups = []
        for i in range(0, count):
            group = FakeGroup.create_one_group(attrs)
            groups.append(group)

        return groups

    @staticmethod
    def get_groups(groups=None, count=2):
        """Get an iterable MagicMock object with a list of faked groups.

        If groups list is provided, then initialize the Mock object with
        the list. Otherwise create one.

        :param List groups:
            A list of FakeResource objects faking groups
        :param Integer count:
            The number of groups to be faked
        :return:
            An iterable Mock object with side_effect set to a list of faked
            groups
        """
        if groups is None:
            groups = FakeGroup.create_groups(count)

        return mock.Mock(side_effect=groups)


class FakeEndpoint:
    """Fake one or more endpoint."""

    @staticmethod
    def create_one_endpoint(attrs=None):
        """Create a fake endpoint.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, url, and so on
        """

        attrs = attrs or {}

        # set default attributes.
        endpoint_info = {
            'id': 'endpoint-id-' + uuid.uuid4().hex,
            'url': 'url-' + uuid.uuid4().hex,
            'region': 'endpoint-region-' + uuid.uuid4().hex,
            'interface': 'admin',
            'service_id': 'service-id-' + uuid.uuid4().hex,
            'enabled': True,
            'links': 'links-' + uuid.uuid4().hex,
        }
        endpoint_info.update(attrs)

        endpoint = fakes.FakeResource(
            info=copy.deepcopy(endpoint_info), loaded=True
        )
        return endpoint

    @staticmethod
    def create_one_endpoint_filter(attrs=None):
        """Create a fake endpoint project relationship.

        :param Dictionary attrs:
            A dictionary with all attributes of endpoint filter
        :return:
            A FakeResource object with project, endpoint and so on
        """
        attrs = attrs or {}

        # Set default attribute
        endpoint_filter_info = {
            'project': 'project-id-' + uuid.uuid4().hex,
            'endpoint': 'endpoint-id-' + uuid.uuid4().hex,
        }

        # Overwrite default attributes if there are some attributes set
        endpoint_filter_info.update(attrs)

        endpoint_filter = FakeModel(copy.deepcopy(endpoint_filter_info))

        return endpoint_filter


class FakeEndpointGroup:
    """Fake one or more endpoint group."""

    @staticmethod
    def create_one_endpointgroup(attrs=None):
        """Create a fake endpoint group.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, url, and so on
        """

        attrs = attrs or {}

        # set default attributes.
        endpointgroup_info = {
            'id': 'endpoint-group-id-' + uuid.uuid4().hex,
            'name': 'endpoint-group-name-' + uuid.uuid4().hex,
            'filters': {
                'region': 'region-' + uuid.uuid4().hex,
                'service_id': 'service-id-' + uuid.uuid4().hex,
            },
            'description': 'endpoint-group-description-' + uuid.uuid4().hex,
            'links': 'links-' + uuid.uuid4().hex,
        }
        endpointgroup_info.update(attrs)

        endpoint = fakes.FakeResource(
            info=copy.deepcopy(endpointgroup_info), loaded=True
        )
        return endpoint

    @staticmethod
    def create_one_endpointgroup_filter(attrs=None):
        """Create a fake endpoint project relationship.

        :param Dictionary attrs:
            A dictionary with all attributes of endpointgroup filter
        :return:
            A FakeResource object with project, endpointgroup and so on
        """
        attrs = attrs or {}

        # Set default attribute
        endpointgroup_filter_info = {
            'project': 'project-id-' + uuid.uuid4().hex,
            'endpointgroup': 'endpointgroup-id-' + uuid.uuid4().hex,
        }

        # Overwrite default attributes if there are some attributes set
        endpointgroup_filter_info.update(attrs)

        endpointgroup_filter = FakeModel(
            copy.deepcopy(endpointgroup_filter_info)
        )

        return endpointgroup_filter


class FakeService:
    """Fake one or more service."""

    @staticmethod
    def create_one_service(attrs=None):
        """Create a fake service.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, name, and so on
        """

        attrs = attrs or {}

        # set default attributes.
        service_info = {
            'id': 'service-id-' + uuid.uuid4().hex,
            'name': 'service-name-' + uuid.uuid4().hex,
            'type': 'service-type-' + uuid.uuid4().hex,
            'description': 'service-description-' + uuid.uuid4().hex,
            'enabled': True,
            'links': 'links-' + uuid.uuid4().hex,
        }
        service_info.update(attrs)

        service = fakes.FakeResource(
            info=copy.deepcopy(service_info), loaded=True
        )
        return service


class FakeRoleAssignment:
    """Fake one or more role assignment."""

    @staticmethod
    def create_one_role_assignment(attrs=None):
        """Create a fake role assignment.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with scope, user, and so on
        """

        attrs = attrs or {}

        # set default attributes.
        role_assignment_info = {
            'scope': {'project': {'id': 'project-id-' + uuid.uuid4().hex}},
            'user': {'id': 'user-id-' + uuid.uuid4().hex},
            'role': {'id': 'role-id-' + uuid.uuid4().hex},
        }
        role_assignment_info.update(attrs)

        role_assignment = fakes.FakeResource(
            info=copy.deepcopy(role_assignment_info), loaded=True
        )

        return role_assignment


class FakeImpliedRoleResponse:
    """Fake one or more role assignment."""

    def __init__(self, prior_role, implied_roles):
        self.prior_role = prior_role
        self.implies = [role for role in implied_roles]

    @staticmethod
    def create_list():
        """Create a fake implied role list response.

        :return:
            A list of FakeImpliedRoleResponse objects
        """

        # set default attributes.
        implied_roles = [FakeImpliedRoleResponse(ROLES[0], [ROLES[1]])]

        return implied_roles



#   Copyright 2014 eBay Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v2 Token action implementations"""

from osc_lib import exceptions

from openstackclient import command
from openstackclient.i18n import _


class IssueToken(command.ShowOne):
    _description = _("Issue new token")

    # scoped token is optional
    required_scope = False

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        auth_ref = self.app.client_manager.auth_ref
        if not auth_ref:
            raise exceptions.AuthorizationFailure(
                "Only an authorized user may issue a new token."
            )

        data = {}
        if auth_ref.auth_token:
            data['id'] = auth_ref.auth_token
        if auth_ref.expires:
            datetime_obj = auth_ref.expires
            expires_str = datetime_obj.strftime('%Y-%m-%dT%H:%M:%S%z')
            data['expires'] = expires_str
        if auth_ref.project_id:
            data['project_id'] = auth_ref.project_id
        if auth_ref.user_id:
            data['user_id'] = auth_ref.user_id
        return zip(*sorted(data.items()))


class RevokeToken(command.Command):
    _description = _("Revoke existing token")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'token',
            metavar='<token>',
            help=_('Token to be deleted'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        identity_client.tokens.delete(parsed_args.token)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures

from openstackclient.tests.functional.volume import base


class BaseVolumeTests(base.BaseVolumeTests):
    """Base class for Volume functional tests."""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.haz_volume_v2 = cls.is_service_enabled('block-storage', '2.0')

    def setUp(self):
        super().setUp()

        if not self.haz_volume_v2:
            self.skipTest("No Volume v2 service present")

        ver_fixture = fixtures.EnvironmentVariable(
            'OS_VOLUME_API_VERSION', '2'
        )
        self.useFixture(ver_fixture)



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 Domain action implementations"""

import logging

from openstack import exceptions as sdk_exceptions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common


LOG = logging.getLogger(__name__)


def _format_domain(domain):
    columns = (
        'id',
        'name',
        'is_enabled',
        'description',
        'options',
    )
    column_headers = (
        'id',
        'name',
        'enabled',
        'description',
        'options',
    )

    return (
        column_headers,
        utils.get_item_properties(
            domain,
            columns,
        ),
    )


class CreateDomain(command.ShowOne):
    _description = _("Create new domain")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<domain-name>',
            help=_('New domain name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New domain description'),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            dest='is_enabled',
            action='store_true',
            default=True,
            help=_('Enable domain (default)'),
        )
        enable_group.add_argument(
            '--disable',
            dest='is_enabled',
            action='store_false',
            help=_('Disable domain'),
        )
        parser.add_argument(
            '--or-show',
            action='store_true',
            help=_('Return existing domain'),
        )
        common.add_resource_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        options = {}
        if parsed_args.immutable is not None:
            options['immutable'] = parsed_args.immutable

        try:
            domain = identity_client.create_domain(
                name=parsed_args.name,
                description=parsed_args.description,
                options=options,
                is_enabled=parsed_args.is_enabled,
            )
        except sdk_exceptions.ConflictException:
            if parsed_args.or_show:
                domain = identity_client.find_domain(
                    parsed_args.name, ignore_missing=False
                )
                LOG.info(_('Returning existing domain %s'), domain.name)
            else:
                raise

        return _format_domain(domain)


class DeleteDomain(command.Command):
    _description = _("Delete domain(s)")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            nargs='+',
            help=_('Domain(s) to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity
        result = 0
        for i in parsed_args.domain:
            try:
                domain = identity_client.find_domain(i, ignore_missing=False)
                identity_client.delete_domain(domain.id)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete domain with name or "
                        "ID '%(domain)s': %(e)s"
                    ),
                    {'domain': i, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.domain)
            msg = _("%(result)s of %(total)s domains failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListDomain(command.Lister):
    _description = _("List domains")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('The domain name'),
        )
        parser.add_argument(
            '--enabled',
            dest='is_enabled',
            action='store_true',
            help=_('The domains that are enabled will be returned'),
        )
        return parser

    def take_action(self, parsed_args):
        kwargs = {}
        if parsed_args.name:
            kwargs['name'] = parsed_args.name
        if parsed_args.is_enabled:
            kwargs['is_enabled'] = True

        columns = ('id', 'name', 'is_enabled', 'description')
        column_headers = ('ID', 'Name', 'Enabled', 'Description')
        data = self.app.client_manager.sdk_connection.identity.domains(
            **kwargs
        )

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class SetDomain(command.Command):
    _description = _("Set domain properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help=_('Domain to modify (name or ID)'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New domain name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New domain description'),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            dest='is_enabled',
            action='store_true',
            default=None,
            help=_('Enable domain'),
        )
        enable_group.add_argument(
            '--disable',
            dest='is_enabled',
            action='store_false',
            default=None,
            help=_('Disable domain'),
        )
        common.add_resource_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity
        domain = identity_client.find_domain(
            parsed_args.domain, ignore_missing=False
        )
        kwargs = {}
        if parsed_args.name:
            kwargs['name'] = parsed_args.name
        if parsed_args.description:
            kwargs['description'] = parsed_args.description
        if parsed_args.is_enabled is not None:
            kwargs['is_enabled'] = parsed_args.is_enabled
        if parsed_args.immutable is not None:
            kwargs['options'] = {'immutable': parsed_args.immutable}

        identity_client.update_domain(domain.id, **kwargs)


class ShowDomain(command.ShowOne):
    _description = _("Display domain details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help=_('Domain to display (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity
        domain = identity_client.find_domain(
            parsed_args.domain, ignore_missing=False
        )

        return _format_domain(domain)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from cinderclient import api_versions
from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


def _format_group_type(group):
    columns = (
        'id',
        'name',
        'description',
        'is_public',
        'group_specs',
    )
    column_headers = (
        'ID',
        'Name',
        'Description',
        'Is Public',
        'Properties',
    )

    # TODO(stephenfin): Consider using a formatter for volume_types since it's
    # a list
    return (
        column_headers,
        utils.get_item_properties(
            group,
            columns,
            formatters={
                'group_specs': format_columns.DictColumn,
            },
        ),
    )


class CreateVolumeGroupType(command.ShowOne):
    """Create a volume group type.

    This command requires ``--os-volume-api-version`` 3.11 or greater.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_('Name of new volume group type.'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description of the volume group type.'),
        )
        type_group = parser.add_mutually_exclusive_group()
        type_group.add_argument(
            '--public',
            dest='is_public',
            action='store_true',
            default=True,
            help=_(
                'Volume group type is available to other projects (default)'
            ),
        )
        type_group.add_argument(
            '--private',
            dest='is_public',
            action='store_false',
            help=_('Volume group type is not available to other projects'),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume

        if volume_client.api_version < api_versions.APIVersion('3.11'):
            msg = _(
                "--os-volume-api-version 3.11 or greater is required to "
                "support the 'volume group type create' command"
            )
            raise exceptions.CommandError(msg)

        group_type = volume_client.group_types.create(
            parsed_args.name, parsed_args.description, parsed_args.is_public
        )

        return _format_group_type(group_type)


class DeleteVolumeGroupType(command.Command):
    """Delete a volume group type.

    This command requires ``--os-volume-api-version`` 3.11 or greater.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group_type',
            metavar='<group_type>',
            help=_('Name or ID of volume group type to delete'),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume

        if volume_client.api_version < api_versions.APIVersion('3.11'):
            msg = _(
                "--os-volume-api-version 3.11 or greater is required to "
                "support the 'volume group type delete' command"
            )
            raise exceptions.CommandError(msg)

        group_type = utils.find_resource(
            volume_client.group_types,
            parsed_args.group_type,
        )

        volume_client.group_types.delete(group_type.id)


class SetVolumeGroupType(command.ShowOne):
    """Update a volume group type.

    This command requires ``--os-volume-api-version`` 3.11 or greater.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group_type',
            metavar='<group_type>',
            help=_('Name or ID of volume group type.'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New name for volume group type.'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New description for volume group type.'),
        )
        type_group = parser.add_mutually_exclusive_group()
        type_group.add_argument(
            '--public',
            dest='is_public',
            action='store_true',
            default=None,
            help=_('Make volume group type available to other projects.'),
        )
        type_group.add_argument(
            '--private',
            dest='is_public',
            action='store_false',
            help=_('Make volume group type unavailable to other projects.'),
        )
        parser.add_argument(
            '--no-property',
            action='store_true',
            help=_(
                'Remove all properties from this volume group type '
                '(specify both --no-property and --property '
                'to remove the current properties before setting '
                'new properties)'
            ),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            dest='properties',
            help=_(
                'Property to add or modify for this volume group type '
                '(repeat option to set multiple properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume

        if volume_client.api_version < api_versions.APIVersion('3.11'):
            msg = _(
                "--os-volume-api-version 3.11 or greater is required to "
                "support the 'volume group type set' command"
            )
            raise exceptions.CommandError(msg)

        group_type = utils.find_resource(
            volume_client.group_types,
            parsed_args.group_type,
        )

        kwargs = {}
        errors = 0

        if parsed_args.name is not None:
            kwargs['name'] = parsed_args.name

        if parsed_args.description is not None:
            kwargs['description'] = parsed_args.description

        if parsed_args.is_public is not None:
            kwargs['is_public'] = parsed_args.is_public

        if kwargs:
            try:
                group_type = volume_client.group_types.update(
                    group_type.id, **kwargs
                )
            except Exception as e:
                LOG.error(_("Failed to update group type: %s"), e)
                errors += 1

        if parsed_args.no_property:
            try:
                keys = group_type.get_keys().keys()
                group_type.unset_keys(keys)
            except Exception as e:
                LOG.error(_("Failed to clear group type properties: %s"), e)
                errors += 1

        if parsed_args.properties:
            try:
                group_type.set_keys(parsed_args.properties)
            except Exception as e:
                LOG.error(_("Failed to set group type properties: %s"), e)
                errors += 1

        if errors > 0:
            msg = _("Command Failed: One or more of the operations failed")
            raise exceptions.CommandError()

        return _format_group_type(group_type)


class UnsetVolumeGroupType(command.ShowOne):
    """Unset properties of a volume group type.

    This command requires ``--os-volume-api-version`` 3.11 or greater.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group_type',
            metavar='<group_type>',
            help=_('Name or ID of volume group type.'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            dest='properties',
            help=_(
                'Property to remove from this volume group type '
                '(repeat option to unset multiple properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume

        if volume_client.api_version < api_versions.APIVersion('3.11'):
            msg = _(
                "--os-volume-api-version 3.11 or greater is required to "
                "support the 'volume group type unset' command"
            )
            raise exceptions.CommandError(msg)

        group_type = utils.find_resource(
            volume_client.group_types,
            parsed_args.group_type,
        )

        group_type.unset_keys(parsed_args.properties)

        group_type = utils.find_resource(
            volume_client.group_types,
            parsed_args.group_type,
        )

        return _format_group_type(group_type)


class ListVolumeGroupType(command.Lister):
    """Lists all volume group types.

    This command requires ``--os-volume-api-version`` 3.11 or greater.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--default',
            action='store_true',
            dest='show_default',
            default=False,
            help=_('List the default volume group type.'),
        )
        # TODO(stephenfin): Add once we have an equivalent command for
        # 'cinder list-filters'
        # parser.add_argument(
        #     '--filter',
        #     metavar='<key=value>',
        #     action=parseractions.KeyValueAction,
        #     dest='filters',
        #     help=_(
        #         "Filter key and value pairs. Use 'foo' to "
        #         "check enabled filters from server. Use 'key~=value' for "
        #         "inexact filtering if the key supports "
        #         "(supported by --os-volume-api-version 3.33 or above)"
        #     ),
        # )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume

        if volume_client.api_version < api_versions.APIVersion('3.11'):
            msg = _(
                "--os-volume-api-version 3.11 or greater is required to "
                "support the 'volume group type list' command"
            )
            raise exceptions.CommandError(msg)

        if parsed_args.show_default:
            group_types = [volume_client.group_types.default()]
        else:
            group_types = volume_client.group_types.list()

        column_headers = (
            'ID',
            'Name',
            'Is Public',
            'Properties',
        )
        columns = (
            'id',
            'name',
            'is_public',
            'group_specs',
        )

        return (
            column_headers,
            (utils.get_item_properties(a, columns) for a in group_types),
        )


class ShowVolumeGroupType(command.ShowOne):
    """Show detailed information for a volume group type.

    This command requires ``--os-volume-api-version`` 3.11 or greater.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group_type',
            metavar='<group_type>',
            help=_('Name or ID of volume group type.'),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume

        if volume_client.api_version < api_versions.APIVersion('3.11'):
            msg = _(
                "--os-volume-api-version 3.11 or greater is required to "
                "support the 'volume group type show' command"
            )
            raise exceptions.CommandError(msg)

        group_type = utils.find_resource(
            volume_client.group_types,
            parsed_args.group,
        )

        return _format_group_type(group_type)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils

from openstackclient.tests.functional.identity.v3 import common


class ServiceProviderTests(common.IdentityTests):
    # Introduce functional test cases for command 'Service Provider'

    def test_sp_create(self):
        self._create_dummy_sp(add_clean_up=True)

    def test_sp_delete(self):
        service_provider = self._create_dummy_sp(add_clean_up=False)
        raw_output = self.openstack(
            f'service provider delete {service_provider}'
        )
        self.assertEqual(0, len(raw_output))

    def test_sp_multi_delete(self):
        sp1 = self._create_dummy_sp(add_clean_up=False)
        sp2 = self._create_dummy_sp(add_clean_up=False)
        raw_output = self.openstack(f'service provider delete {sp1} {sp2}')
        self.assertEqual(0, len(raw_output))

    def test_sp_show(self):
        service_provider = self._create_dummy_sp(add_clean_up=True)
        raw_output = self.openstack(
            f'service provider show {service_provider}'
        )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.SERVICE_PROVIDER_FIELDS)

    def test_sp_list(self):
        self._create_dummy_sp(add_clean_up=True)
        raw_output = self.openstack('service provider list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, self.SERVICE_PROVIDER_LIST_HEADERS)

    def test_sp_set(self):
        service_provider = self._create_dummy_sp(add_clean_up=True)
        new_description = data_utils.rand_name('newDescription')
        raw_output = self.openstack(
            f'service provider set '
            f'{service_provider} '
            f'--description {new_description}'
        )
        updated_value = self.parse_show_as_object(raw_output)
        self.assertEqual(new_description, updated_value.get('description'))



#   Copyright 2016 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


from openstack.compute.v2 import server as _server
from openstack.test import fakes as sdk_fakes

from openstackclient.compute.v2 import console
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes
from openstackclient.tests.unit import utils


class TestConsoleLog(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self._server = sdk_fakes.generate_fake_resource(_server.Server)
        self.compute_client.find_server.return_value = self._server

        self.cmd = console.ShowConsoleLog(self.app, None)

    def test_show_no_args(self):
        arglist = []
        verifylist = []
        self.assertRaises(
            utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show(self):
        arglist = ['fake_server']
        verifylist = [('server', 'fake_server')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        output = {'output': '1st line\n2nd line\n'}
        self.compute_client.get_server_console_output.return_value = output
        self.cmd.take_action(parsed_args)

        self.compute_client.find_server.assert_called_with(
            name_or_id='fake_server', ignore_missing=False
        )
        self.compute_client.get_server_console_output.assert_called_with(
            self._server.id, length=None
        )
        stdout = self.app.stdout.content
        self.assertEqual(stdout[0], output['output'])

    def test_show_lines(self):
        arglist = ['fake_server', '--lines', '15']
        verifylist = [('server', 'fake_server'), ('lines', 15)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        output = {'output': '1st line\n2nd line'}
        self.compute_client.get_server_console_output.return_value = output
        self.cmd.take_action(parsed_args)

        self.compute_client.find_server.assert_called_with(
            name_or_id='fake_server', ignore_missing=False
        )
        self.compute_client.get_server_console_output.assert_called_with(
            self._server.id, length=15
        )


class TestConsoleUrlShow(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self._server = sdk_fakes.generate_fake_resource(_server.Server)
        self.compute_client.find_server.return_value = self._server

        fake_console_data = {
            'url': 'http://localhost',
            'protocol': 'fake_protocol',
            'type': 'fake_type',
        }
        self.compute_client.create_console.return_value = fake_console_data

        self.columns = (
            'protocol',
            'type',
            'url',
        )
        self.data = (
            fake_console_data['protocol'],
            fake_console_data['type'],
            fake_console_data['url'],
        )

        self.cmd = console.ShowConsoleURL(self.app, None)

    def test_console_url_show_by_default(self):
        arglist = [
            'foo_vm',
        ]
        verifylist = [
            ('url_type', 'novnc'),
            ('server', 'foo_vm'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.create_console.assert_called_once_with(
            self._server.id, console_type='novnc'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_console_url_show_with_novnc(self):
        arglist = [
            '--novnc',
            'foo_vm',
        ]
        verifylist = [
            ('url_type', 'novnc'),
            ('server', 'foo_vm'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.create_console.assert_called_once_with(
            self._server.id, console_type='novnc'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_console_url_show_with_xvpvnc(self):
        arglist = [
            '--xvpvnc',
            'foo_vm',
        ]
        verifylist = [
            ('url_type', 'xvpvnc'),
            ('server', 'foo_vm'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.create_console.assert_called_once_with(
            self._server.id, console_type='xvpvnc'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_console_url_show_with_spice_html5(self):
        arglist = [
            '--spice',
            'foo_vm',
        ]
        verifylist = [
            ('url_type', 'spice-html5'),
            ('server', 'foo_vm'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.create_console.assert_called_once_with(
            self._server.id, console_type='spice-html5'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_console_url_show_with_spice_direct(self):
        arglist = [
            '--spice-direct',
            'foo_vm',
        ]
        verifylist = [
            ('url_type', 'spice-direct'),
            ('server', 'foo_vm'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.create_console.assert_called_once_with(
            self._server.id, console_type='spice-direct'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_console_url_show_with_rdp(self):
        arglist = [
            '--rdp',
            'foo_vm',
        ]
        verifylist = [
            ('url_type', 'rdp-html5'),
            ('server', 'foo_vm'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.create_console.assert_called_once_with(
            self._server.id, console_type='rdp-html5'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_console_url_show_with_serial(self):
        arglist = [
            '--serial',
            'foo_vm',
        ]
        verifylist = [
            ('url_type', 'serial'),
            ('server', 'foo_vm'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.create_console.assert_called_once_with(
            self._server.id, console_type='serial'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_console_url_show_with_mks(self):
        arglist = [
            '--mks',
            'foo_vm',
        ]
        verifylist = [
            ('url_type', 'webmks'),
            ('server', 'foo_vm'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.compute_client.create_console.assert_called_once_with(
            self._server.id, console_type='webmks'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#   Copyright 2016 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Compute v2 Server Group action implementations"""

import logging
import typing as ty

from cliff import columns
from openstack import utils as sdk_utils
from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


_formatters: dict[str, type[columns.FormattableColumn[ty.Any]]] = {
    'member_ids': format_columns.ListColumn,
    'policies': format_columns.ListColumn,
    'rules': format_columns.DictColumn,
}


def _get_server_group_columns(item, client):
    column_map = {'member_ids': 'members'}
    hidden_columns = ['metadata', 'location']

    if sdk_utils.supports_microversion(client, '2.64'):
        hidden_columns.append('policies')
    else:
        hidden_columns.append('policy')
        hidden_columns.append('rules')

    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


class CreateServerGroup(command.ShowOne):
    _description = _("Create a new server group.")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_("New server group name"),
        )
        parser.add_argument(
            '--policy',
            metavar='<policy>',
            choices=[
                'affinity',
                'anti-affinity',
                'soft-affinity',
                'soft-anti-affinity',
            ],
            default='affinity',
            help=_(
                "Add a policy to <name> "
                "Specify --os-compute-api-version 2.15 or higher for the "
                "'soft-affinity' or 'soft-anti-affinity' policy."
            ),
        )
        parser.add_argument(
            '--rule',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            default={},
            dest='rules',
            help=_(
                "A rule for the policy. Currently, only the "
                "'max_server_per_host' rule is supported for the "
                "'anti-affinity' policy."
            ),
        )
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.compute

        if parsed_args.policy in ('soft-affinity', 'soft-anti-affinity'):
            if not sdk_utils.supports_microversion(compute_client, '2.15'):
                msg = _(
                    '--os-compute-api-version 2.15 or greater is required to '
                    'support the %s policy'
                )
                raise exceptions.CommandError(msg % parsed_args.policy)

        if parsed_args.rules:
            if not sdk_utils.supports_microversion(compute_client, '2.64'):
                msg = _(
                    '--os-compute-api-version 2.64 or greater is required to '
                    'support the --rule option'
                )
                raise exceptions.CommandError(msg)

        if not sdk_utils.supports_microversion(compute_client, '2.64'):
            kwargs = {
                'name': parsed_args.name,
                'policies': [parsed_args.policy],
            }
        else:
            kwargs = {
                'name': parsed_args.name,
                'policy': parsed_args.policy,
            }

            if parsed_args.rules:
                kwargs['rules'] = parsed_args.rules

        server_group = compute_client.create_server_group(**kwargs)

        display_columns, columns = _get_server_group_columns(
            server_group,
            compute_client,
        )
        data = utils.get_item_properties(
            server_group,
            columns,
            formatters=_formatters,
        )
        return display_columns, data


class DeleteServerGroup(command.Command):
    _description = _("Delete existing server group(s).")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server_group',
            metavar='<server-group>',
            nargs='+',
            help=_("server group(s) to delete (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.compute
        result = 0
        for group in parsed_args.server_group:
            try:
                group_obj = compute_client.find_server_group(
                    group, ignore_missing=False
                )
                compute_client.delete_server_group(group_obj.id)
            # Catch all exceptions in order to avoid to block the next deleting
            except Exception as e:
                result += 1
                LOG.error(e)

        if result > 0:
            total = len(parsed_args.server_group)
            msg = _("%(result)s of %(total)s server groups failed to delete.")
            raise exceptions.CommandError(
                msg % {"result": result, "total": total}
            )


class ListServerGroup(command.Lister):
    _description = _("List all server groups.")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            action='store_true',
            default=False,
            help=_("Display information from all projects (admin only)"),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_("List additional fields in output"),
        )
        # TODO(stephenfin): This should really be a --marker option, but alas
        # the API doesn't support that for some reason
        pagination.add_offset_pagination_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.compute

        kwargs = {}

        if parsed_args.all_projects:
            kwargs['all_projects'] = parsed_args.all_projects

        if parsed_args.offset:
            kwargs['offset'] = parsed_args.offset

        if parsed_args.limit:
            kwargs['limit'] = parsed_args.limit

        data = compute_client.server_groups(**kwargs)

        policy_key = 'Policies'
        if sdk_utils.supports_microversion(compute_client, '2.64'):
            policy_key = 'Policy'

        columns: tuple[str, ...] = (
            'id',
            'name',
            policy_key.lower(),
        )
        column_headers: tuple[str, ...] = (
            'ID',
            'Name',
            policy_key,
        )
        if parsed_args.long:
            columns += (
                'member_ids',
                'project_id',
                'user_id',
            )
            column_headers += (
                'Members',
                'Project Id',
                'User Id',
            )

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters=_formatters,
                )
                for s in data
            ),
        )


class ShowServerGroup(command.ShowOne):
    _description = _("Display server group details.")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server_group',
            metavar='<server-group>',
            help=_("server group to display (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.compute
        group = compute_client.find_server_group(
            parsed_args.server_group, ignore_missing=False
        )
        display_columns, columns = _get_server_group_columns(
            group,
            compute_client,
        )
        data = utils.get_item_properties(
            group, columns, formatters=_formatters
        )
        return display_columns, data



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest.mock import call

from osc_lib import exceptions

from openstackclient.network.v2 import address_group
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes_v3
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestAddressGroup(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        # Get a shortcut to the ProjectManager Mock
        self.projects_mock = self.identity_client.projects
        # Get a shortcut to the DomainManager Mock
        self.domains_mock = self.identity_client.domains


class TestCreateAddressGroup(TestAddressGroup):
    project = identity_fakes_v3.FakeProject.create_one_project()
    domain = identity_fakes_v3.FakeDomain.create_one_domain()
    # The new address group created.
    new_address_group = network_fakes.create_one_address_group(
        attrs={
            'project_id': project.id,
        }
    )
    columns = (
        'addresses',
        'description',
        'id',
        'name',
        'project_id',
    )
    data = (
        new_address_group.addresses,
        new_address_group.description,
        new_address_group.id,
        new_address_group.name,
        new_address_group.project_id,
    )

    def setUp(self):
        super().setUp()
        self.network_client.create_address_group.return_value = (
            self.new_address_group
        )

        # Get the command object to test
        self.cmd = address_group.CreateAddressGroup(self.app, None)

        self.projects_mock.get.return_value = self.project
        self.domains_mock.get.return_value = self.domain

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_default_options(self):
        arglist = [
            self.new_address_group.name,
        ]
        verifylist = [
            ('project', None),
            ('name', self.new_address_group.name),
            ('description', None),
            ('address', []),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_address_group.assert_called_once_with(
            **{
                'name': self.new_address_group.name,
                'addresses': [],
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_all_options(self):
        arglist = [
            '--project',
            self.project.name,
            '--project-domain',
            self.domain.name,
            '--address',
            '10.0.0.1',
            '--description',
            self.new_address_group.description,
            self.new_address_group.name,
        ]
        verifylist = [
            ('project', self.project.name),
            ('project_domain', self.domain.name),
            ('address', ['10.0.0.1']),
            ('description', self.new_address_group.description),
            ('name', self.new_address_group.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_address_group.assert_called_once_with(
            **{
                'addresses': ['10.0.0.1/32'],
                'project_id': self.project.id,
                'name': self.new_address_group.name,
                'description': self.new_address_group.description,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestDeleteAddressGroup(TestAddressGroup):
    # The address group to delete.
    _address_groups = network_fakes.create_address_groups(count=2)

    def setUp(self):
        super().setUp()
        self.network_client.delete_address_group.return_value = None
        self.network_client.find_address_group = (
            network_fakes.get_address_groups(
                address_groups=self._address_groups
            )
        )

        # Get the command object to test
        self.cmd = address_group.DeleteAddressGroup(self.app, None)

    def test_address_group_delete(self):
        arglist = [
            self._address_groups[0].name,
        ]
        verifylist = [
            ('address_group', [self._address_groups[0].name]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.find_address_group.assert_called_once_with(
            self._address_groups[0].name, ignore_missing=False
        )
        self.network_client.delete_address_group.assert_called_once_with(
            self._address_groups[0]
        )
        self.assertIsNone(result)

    def test_multi_address_groups_delete(self):
        arglist = []

        for a in self._address_groups:
            arglist.append(a.name)
        verifylist = [
            ('address_group', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = []
        for a in self._address_groups:
            calls.append(call(a))
        self.network_client.delete_address_group.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_multi_address_groups_delete_with_exception(self):
        arglist = [
            self._address_groups[0].name,
            'unexist_address_group',
        ]
        verifylist = [
            (
                'address_group',
                [self._address_groups[0].name, 'unexist_address_group'],
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [self._address_groups[0], exceptions.CommandError]
        self.network_client.find_address_group.side_effect = find_mock_result

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 address groups failed to delete.', str(e))

        self.network_client.find_address_group.assert_any_call(
            self._address_groups[0].name, ignore_missing=False
        )
        self.network_client.find_address_group.assert_any_call(
            'unexist_address_group', ignore_missing=False
        )
        self.network_client.delete_address_group.assert_called_once_with(
            self._address_groups[0]
        )


class TestListAddressGroup(TestAddressGroup):
    # The address groups to list up.
    address_groups = network_fakes.create_address_groups(count=3)
    columns = (
        'ID',
        'Name',
        'Description',
        'Project',
        'Addresses',
    )
    data = []
    for group in address_groups:
        data.append(
            (
                group.id,
                group.name,
                group.description,
                group.project_id,
                group.addresses,
            )
        )

    def setUp(self):
        super().setUp()
        self.network_client.address_groups.return_value = self.address_groups

        # Get the command object to test
        self.cmd = address_group.ListAddressGroup(self.app, None)

    def test_address_group_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.address_groups.assert_called_once_with(**{})
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_address_group_list_name(self):
        arglist = [
            '--name',
            self.address_groups[0].name,
        ]
        verifylist = [
            ('name', self.address_groups[0].name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.address_groups.assert_called_once_with(
            **{'name': self.address_groups[0].name}
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_address_group_list_project(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        self.projects_mock.get.return_value = project
        arglist = [
            '--project',
            project.id,
        ]
        verifylist = [
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.address_groups.assert_called_once_with(
            project_id=project.id
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_address_group_project_domain(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        self.projects_mock.get.return_value = project
        arglist = [
            '--project',
            project.id,
            '--project-domain',
            project.domain_id,
        ]
        verifylist = [
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.address_groups.assert_called_once_with(
            project_id=project.id
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))


class TestSetAddressGroup(TestAddressGroup):
    # The address group to set.
    _address_group = network_fakes.create_one_address_group()

    def setUp(self):
        super().setUp()
        self.network_client.update_address_group.return_value = None
        self.network_client.find_address_group.return_value = (
            self._address_group
        )

        self.network_client.add_addresses_to_address_group.return_value = (
            self._address_group
        )

        # Get the command object to test
        self.cmd = address_group.SetAddressGroup(self.app, None)

    def test_set_nothing(self):
        arglist = [
            self._address_group.name,
        ]
        verifylist = [
            ('address_group', self._address_group.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_address_group.assert_not_called()
        self.network_client.add_addresses_to_address_group.assert_not_called()
        self.assertIsNone(result)

    def test_set_name_and_description(self):
        arglist = [
            '--name',
            'new_address_group_name',
            '--description',
            'new_address_group_description',
            self._address_group.name,
        ]
        verifylist = [
            ('name', 'new_address_group_name'),
            ('description', 'new_address_group_description'),
            ('address_group', self._address_group.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'name': "new_address_group_name",
            'description': 'new_address_group_description',
        }
        self.network_client.update_address_group.assert_called_with(
            self._address_group, **attrs
        )
        self.assertIsNone(result)

    def test_set_one_address(self):
        arglist = [
            self._address_group.name,
            '--address',
            '10.0.0.2',
        ]
        verifylist = [
            ('address_group', self._address_group.name),
            ('address', ['10.0.0.2']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.add_addresses_to_address_group.assert_called_once_with(
            self._address_group, ['10.0.0.2/32']
        )
        self.assertIsNone(result)

    def test_set_multiple_addresses(self):
        arglist = [
            self._address_group.name,
            '--address',
            '10.0.0.2',
            '--address',
            '2001::/16',
        ]
        verifylist = [
            ('address_group', self._address_group.name),
            ('address', ['10.0.0.2', '2001::/16']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.add_addresses_to_address_group.assert_called_once_with(
            self._address_group, ['10.0.0.2/32', '2001::/16']
        )
        self.assertIsNone(result)


class TestShowAddressGroup(TestAddressGroup):
    # The address group to show.
    _address_group = network_fakes.create_one_address_group()
    columns = (
        'addresses',
        'description',
        'id',
        'name',
        'project_id',
    )
    data = (
        _address_group.addresses,
        _address_group.description,
        _address_group.id,
        _address_group.name,
        _address_group.project_id,
    )

    def setUp(self):
        super().setUp()
        self.network_client.find_address_group.return_value = (
            self._address_group
        )

        # Get the command object to test
        self.cmd = address_group.ShowAddressGroup(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show_all_options(self):
        arglist = [
            self._address_group.name,
        ]
        verifylist = [
            ('address_group', self._address_group.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.find_address_group.assert_called_once_with(
            self._address_group.name, ignore_missing=False
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))


class TestUnsetAddressGroup(TestAddressGroup):
    # The address group to unset.
    _address_group = network_fakes.create_one_address_group()

    def setUp(self):
        super().setUp()
        self.network_client.find_address_group.return_value = (
            self._address_group
        )

        self.network_client.remove_addresses_from_address_group.return_value = self._address_group

        # Get the command object to test
        self.cmd = address_group.UnsetAddressGroup(self.app, None)

    def test_unset_nothing(self):
        arglist = [
            self._address_group.name,
        ]
        verifylist = [
            ('address_group', self._address_group.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.remove_addresses_from_address_group.assert_not_called()
        self.assertIsNone(result)

    def test_unset_one_address(self):
        arglist = [
            self._address_group.name,
            '--address',
            '10.0.0.2',
        ]
        verifylist = [
            ('address_group', self._address_group.name),
            ('address', ['10.0.0.2']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.remove_addresses_from_address_group.assert_called_once_with(  # noqa: E501
            self._address_group, ['10.0.0.2/32']
        )
        self.assertIsNone(result)

    def test_unset_multiple_addresses(self):
        arglist = [
            self._address_group.name,
            '--address',
            '10.0.0.2',
            '--address',
            '2001::/16',
        ]
        verifylist = [
            ('address_group', self._address_group.name),
            ('address', ['10.0.0.2', '2001::/16']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.network_client.remove_addresses_from_address_group.assert_called_once_with(  # noqa: E501
            self._address_group, ['10.0.0.2/32', '2001::/16']
        )
        self.assertIsNone(result)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random

from openstackclient.tests.functional.network.v2 import common


class SecurityGroupRuleTests(common.NetworkTests):
    """Functional tests for security group rule"""

    def setUp(self):
        super().setUp()

        if not self.is_extension_enabled("security-groups-default-rules"):
            self.skipTest("No security-groups-default-rules extension present")

        self.port = random.randint(1, 65535)
        self.protocol = random.choice(["tcp", "udp"])
        self.direction = random.choice(["ingress", "egress"])
        # Create the default security group rule.
        cmd_output = self.openstack(
            'default security group rule create '
            f'--protocol {self.protocol} '
            f'--dst-port {self.port}:{self.port} '
            f'--{self.direction} --ethertype IPv4 ',
            parse_output=True,
        )
        self.addCleanup(
            self.openstack,
            'default security group rule delete ' + cmd_output['id'],
        )
        self.DEFAULT_SG_RULE_ID = cmd_output['id']

    def test_security_group_rule_list(self):
        cmd_output = self.openstack(
            'default security group rule list ',
            parse_output=True,
        )
        self.assertIn(
            self.DEFAULT_SG_RULE_ID, [rule['ID'] for rule in cmd_output]
        )

    def test_security_group_rule_show(self):
        cmd_output = self.openstack(
            'default security group rule show ' + self.DEFAULT_SG_RULE_ID,
            parse_output=True,
        )
        self.assertEqual(self.DEFAULT_SG_RULE_ID, cmd_output['id'])
        self.assertEqual(self.protocol, cmd_output['protocol'])
        self.assertEqual(self.port, cmd_output['port_range_min'])
        self.assertEqual(self.port, cmd_output['port_range_max'])
        self.assertEqual(self.direction, cmd_output['direction'])



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Image v2 API Library"""

from openstackclient.api import image_v1


class APIv2(image_v1.APIv1):
    """Image v2 API"""

    _endpoint_suffix = '/v2'

    def _munge_url(self):
        # Hack this until discovery is up, and ignore parent endpoint setting
        if not self.endpoint.endswith(self._endpoint_suffix):
            self.endpoint = self.endpoint + self._endpoint_suffix

    def image_list(
        self,
        detailed=False,
        public=False,
        private=False,
        community=False,
        shared=False,
        **filter,
    ):
        """Get available images

        can add limit/marker

        :param detailed:
            For v1 compatibility only, ignored as v2 is always 'detailed'
        :param public:
            Return public images if True
        :param private:
            Return private images if True
        :param community:
            Return commuity images if True
        :param shared:
            Return shared images if True

        If public, private, community and shared are all True or all False
        then all images are returned.  All arguments False is equivalent to no
        filter and all images are returned.  All arguments True is a filter
        that includes all public, private, community and shared images which
        is the same set as all images.

        http://docs.openstack.org/api/openstack-image-service/2.0/content/list-images.html
        """

        if not public and not private and not community and not shared:
            # No filtering for all False
            filter.pop('visibility', None)
        elif public:
            filter['visibility'] = 'public'
        elif private:
            filter['visibility'] = 'private'
        elif community:
            filter['visibility'] = 'community'
        elif shared:
            filter['visibility'] = 'shared'

        url = "/images"
        if detailed:
            # Because we can't all use /details
            url += "/detail"

        return self.list(url, **filter)['images']



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v2 Assignment action implementations"""

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _  # noqa


class ListRoleAssignment(command.Lister):
    _description = _("List role assignments")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--user',
            metavar='<user>',
            help='User to filter (name or ID)',
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help='Project to filter (name or ID)',
        )
        parser.add_argument(
            '--names',
            action="store_true",
            help='Display names instead of IDs',
        )
        parser.add_argument(
            '--auth-user',
            action="store_true",
            dest='authuser',
            help='Only list assignments for the authenticated user',
        )
        parser.add_argument(
            '--auth-project',
            action="store_true",
            dest='authproject',
            help='Only list assignments for the project to which the '
            'authenticated user\'s token is scoped',
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        auth_ref = self.app.client_manager.auth_ref

        include_names = True if parsed_args.names else False

        user = None
        if parsed_args.user:
            user = utils.find_resource(
                identity_client.users,
                parsed_args.user,
            )
        elif parsed_args.authuser:
            if auth_ref and auth_ref.user_id:
                user = utils.find_resource(
                    identity_client.users, auth_ref.user_id
                )

        project = None
        if parsed_args.project:
            project = utils.find_resource(
                identity_client.projects,
                parsed_args.project,
            )
        elif parsed_args.authproject:
            if auth_ref and auth_ref.project_id:
                project = utils.find_resource(
                    identity_client.projects, auth_ref.project_id
                )

        # If user or project is not specified, we would ideally list all
        # relevant assignments in the system (to be compatible with v3).
        # However, there is no easy way of doing that in v2.
        if not user or not project:
            msg = _("Project and User must be specified")
            raise exceptions.CommandError(msg)
        else:
            data = identity_client.roles.roles_for_user(user.id, project.id)

        columns = ('Role', 'User', 'Project')
        for user_role in data:
            if include_names:
                setattr(user_role, 'role', user_role.name)
                user_role.user = user.name
                user_role.project = project.name
            else:
                setattr(user_role, 'role', user_role.id)
                user_role.user = user.id
                user_role.project = project.id

        return (
            columns,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys


class _ProgressBarBase:
    """A progress bar provider for a wrapped object.

    Base abstract class used by specific class wrapper to show
    a progress bar when the wrapped object are consumed.

    :param wrapped: Object to wrap that hold data to be consumed.
    :param totalsize: The total size of the data in the wrapped object.

    :note: The progress will be displayed only if sys.stdout is a tty.
    """

    def __init__(self, wrapped, totalsize):
        self._wrapped = wrapped
        self._totalsize = float(totalsize)
        self._show_progress = sys.stdout.isatty() and self._totalsize != 0
        self._percent = 0

    def _display_progress_bar(self, size_read):
        if self._show_progress:
            self._percent += size_read / self._totalsize
            # Output something like this: [==========>             ] 49%
            sys.stdout.write(
                '\r[{:<30}] {:.0%}'.format(
                    '=' * int(round(self._percent * 29)) + '>', self._percent
                )
            )
            sys.stdout.flush()

    def __getattr__(self, attr):
        # Forward other attribute access to the wrapped object.
        return getattr(self._wrapped, attr)


class VerboseFileWrapper(_ProgressBarBase):
    """A file wrapper with a progress bar.

    The file wrapper shows and advances a progress bar whenever the
    wrapped file's read method is called.
    """

    def read(self, *args, **kwargs):
        data = self._wrapped.read(*args, **kwargs)
        if data:
            self._display_progress_bar(len(data))
        else:
            if self._show_progress:
                # Break to a new line from the progress bar for incoming
                # output.
                sys.stdout.write('\n')
        return data



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from cinderclient.v3 import volume_snapshots
from cinderclient.v3 import volumes
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.tests.unit import utils as test_utils

ID = '1after909'
NAME = 'PhilSpector'


class TestFindResourceVolumes(test_utils.TestCase):
    def setUp(self):
        super().setUp()
        api = mock.Mock()
        api.client = mock.Mock()
        api.client.get = mock.Mock()
        resp = mock.Mock()
        body = {"volumes": [{"id": ID, 'name': NAME}]}
        api.client.get.side_effect = [Exception("Not found"), (resp, body)]
        self.manager = volumes.VolumeManager(api)

    def test_find(self):
        result = utils.find_resource(self.manager, NAME)
        self.assertEqual(ID, result.id)
        self.assertEqual(NAME, result.name)

    def test_not_find(self):
        self.assertRaises(
            exceptions.CommandError,
            utils.find_resource,
            self.manager,
            'GeorgeMartin',
        )


class TestFindResourceVolumeSnapshots(test_utils.TestCase):
    def setUp(self):
        super().setUp()
        api = mock.Mock()
        api.client = mock.Mock()
        api.client.get = mock.Mock()
        resp = mock.Mock()
        body = {"snapshots": [{"id": ID, 'name': NAME}]}
        api.client.get.side_effect = [Exception("Not found"), (resp, body)]
        self.manager = volume_snapshots.SnapshotManager(api)

    def test_find(self):
        result = utils.find_resource(self.manager, NAME)
        self.assertEqual(ID, result.id)
        self.assertEqual(NAME, result.name)

    def test_not_find(self):
        self.assertRaises(
            exceptions.CommandError,
            utils.find_resource,
            self.manager,
            'GeorgeMartin',
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import uuid

from openstackclient.tests.functional.network.v2 import common


class L3ConntrackHelperTests(common.NetworkTests):
    def setUp(self):
        super().setUp()

        if not self.is_extension_enabled('l3-conntrack-helper'):
            self.skipTest("No l3-conntrack-helper extension present")

        if not self.is_extension_enabled('expose-l3-conntrack-helper'):
            self.skipTest("No expose-l3-conntrack-helper extension present")

    def _create_router(self):
        router_name = uuid.uuid4().hex
        json_output = self.openstack(
            'router create ' + router_name,
            parse_output=True,
        )
        self.assertIsNotNone(json_output['id'])
        router_id = json_output['id']
        self.addCleanup(self.openstack, 'router delete ' + router_id)
        return router_id

    def _create_helpers(self, router_id, helpers):
        created_helpers = []
        for helper in helpers:
            output = self.openstack(
                'network l3 conntrack helper create {router} '
                '--helper {helper} --protocol {protocol} '
                '--port {port} '.format(
                    router=router_id,
                    helper=helper['helper'],
                    protocol=helper['protocol'],
                    port=helper['port'],
                ),
                parse_output=True,
            )
            self.assertEqual(helper['helper'], output['helper'])
            self.assertEqual(helper['protocol'], output['protocol'])
            self.assertEqual(helper['port'], output['port'])
            created_helpers.append(output)
        return created_helpers

    def test_l3_conntrack_helper_create_and_delete(self):
        """Test create, delete multiple"""

        helpers = [
            {'helper': 'tftp', 'protocol': 'udp', 'port': 69},
            {'helper': 'ftp', 'protocol': 'tcp', 'port': 21},
        ]
        router_id = self._create_router()
        created_helpers = self._create_helpers(router_id, helpers)
        ct_ids = " ".join([ct['id'] for ct in created_helpers])

        raw_output = self.openstack(
            f'--debug network l3 conntrack helper delete {router_id} {ct_ids}'
        )
        self.assertOutput('', raw_output)

    def test_l3_conntrack_helper_list(self):
        helpers = [
            {'helper': 'tftp', 'protocol': 'udp', 'port': 69},
            {'helper': 'ftp', 'protocol': 'tcp', 'port': 21},
        ]
        expected_helpers = [
            {'Helper': 'tftp', 'Protocol': 'udp', 'Port': 69},
            {'Helper': 'ftp', 'Protocol': 'tcp', 'Port': 21},
        ]
        router_id = self._create_router()
        self._create_helpers(router_id, helpers)
        output = self.openstack(
            f'network l3 conntrack helper list {router_id} ',
            parse_output=True,
        )
        for ct in output:
            self.assertEqual(router_id, ct.pop('Router ID'))
            ct.pop("ID")
            self.assertIn(ct, expected_helpers)

    def test_l3_conntrack_helper_set_and_show(self):
        helper = 'tftp'
        proto = 'udp'
        port = 69
        router_id = self._create_router()
        created_helper = self._create_helpers(
            router_id,
            [{'helper': helper, 'protocol': proto, 'port': port}],
        )[0]
        output = self.openstack(
            'network l3 conntrack helper show {router_id} {ct_id} '
            '-f json'.format(
                router_id=router_id,
                ct_id=created_helper['id'],
            ),
            parse_output=True,
        )
        self.assertEqual(port, output['port'])
        self.assertEqual(helper, output['helper'])
        self.assertEqual(proto, output['protocol'])

        raw_output = self.openstack(
            'network l3 conntrack helper set {router_id} {ct_id} '
            '--port {port} '.format(
                router_id=router_id,
                ct_id=created_helper['id'],
                port=port + 1,
            )
        )
        self.assertOutput('', raw_output)

        output = self.openstack(
            'network l3 conntrack helper show {router_id} {ct_id} '
            '-f json'.format(
                router_id=router_id,
                ct_id=created_helper['id'],
            ),
            parse_output=True,
        )
        self.assertEqual(port + 1, output['port'])
        self.assertEqual(helper, output['helper'])
        self.assertEqual(proto, output['protocol'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import fixtures
from tempest.lib.common.utils import data_utils

from openstackclient.tests.functional import base


BASIC_LIST_HEADERS = ['ID', 'Name']
SYSTEM_CLOUD = os.environ.get('OS_SYSTEM_CLOUD', 'devstack-system-admin')


class IdentityTests(base.TestCase):
    """Functional tests for Identity commands."""

    DOMAIN_FIELDS = ['description', 'enabled', 'id', 'name']
    GROUP_FIELDS = ['description', 'domain_id', 'id', 'name']
    TOKEN_FIELDS = ['expires', 'id', 'project_id', 'user_id']
    USER_FIELDS = [
        'email',
        'enabled',
        'id',
        'name',
        'domain_id',
        'default_project_id',
        'description',
        'password_expires_at',
    ]
    PROJECT_FIELDS = [
        'description',
        'id',
        'domain_id',
        'is_domain',
        'enabled',
        'name',
        'parent_id',
    ]
    ROLE_FIELDS = ['id', 'name', 'domain_id', 'description']
    SERVICE_FIELDS = ['id', 'enabled', 'name', 'type', 'description']
    REGION_FIELDS = ['description', 'parent_region', 'region']
    ENDPOINT_FIELDS = [
        'id',
        'region',
        'region_id',
        'service_id',
        'service_name',
        'service_type',
        'enabled',
        'interface',
        'url',
    ]

    REGION_LIST_HEADERS = ['Region', 'Parent Region', 'Description']
    ENDPOINT_LIST_HEADERS = [
        'ID',
        'Region',
        'Service Name',
        'Service Type',
        'Enabled',
        'Interface',
        'URL',
    ]
    ENDPOINT_LIST_PROJECT_HEADERS = ['ID', 'Name']

    IDENTITY_PROVIDER_FIELDS = [
        'description',
        'enabled',
        'id',
        'remote_ids',
        'domain_id',
    ]
    IDENTITY_PROVIDER_LIST_HEADERS = ['ID', 'Enabled', 'Description']

    SERVICE_PROVIDER_FIELDS = [
        'auth_url',
        'description',
        'enabled',
        'id',
        'relay_state_prefix',
        'sp_url',
    ]
    SERVICE_PROVIDER_LIST_HEADERS = [
        'ID',
        'Enabled',
        'Description',
        'Auth URL',
    ]
    IMPLIED_ROLE_LIST_HEADERS = [
        'Prior Role ID',
        'Prior Role Name',
        'Implied Role ID',
        'Implied Role Name',
    ]
    ROLE_ASSIGNMENT_LIST_HEADERS = [
        'Role',
        'User',
        'Group',
        'Project',
        'Domain',
        'System',
        'Inherited',
    ]
    REGISTERED_LIMIT_FIELDS = [
        'id',
        'service_id',
        'resource_name',
        'default_limit',
        'description',
        'region_id',
    ]
    REGISTERED_LIMIT_LIST_HEADERS = [
        'ID',
        'Service ID',
        'Resource Name',
        'Default Limit',
        'Description',
        'Region ID',
    ]
    LIMIT_FIELDS = [
        'id',
        'project_id',
        'service_id',
        'resource_name',
        'resource_limit',
        'description',
        'region_id',
    ]
    LIMIT_LIST_HEADERS = [
        'ID',
        'Project ID',
        'Service ID',
        'Resource Name',
        'Resource Limit',
        'Description',
        'Region ID',
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # create dummy domain
        cls.domain_name = data_utils.rand_name('TestDomain')
        cls.domain_description = data_utils.rand_name('description')
        cls.openstack(
            '--os-identity-api-version 3 '
            'domain create '
            f'--description {cls.domain_description} '
            '--enable '
            f'{cls.domain_name}'
        )

        # create dummy project
        cls.project_name = data_utils.rand_name('TestProject')
        cls.project_description = data_utils.rand_name('description')
        cls.openstack(
            '--os-identity-api-version 3 '
            'project create '
            f'--domain {cls.domain_name} '
            f'--description {cls.project_description} '
            '--enable '
            f'{cls.project_name}'
        )

    @classmethod
    def tearDownClass(cls):
        try:
            # delete dummy project
            cls.openstack(
                '--os-identity-api-version 3 '
                f'project delete {cls.project_name}'
            )
            # disable and delete dummy domain
            cls.openstack(
                '--os-identity-api-version 3 '
                f'domain set --disable {cls.domain_name}'
            )
            cls.openstack(
                f'--os-identity-api-version 3 domain delete {cls.domain_name}'
            )
        finally:
            super().tearDownClass()

    def setUp(self):
        super().setUp()
        # prepare v3 env
        ver_fixture = fixtures.EnvironmentVariable(
            'OS_IDENTITY_API_VERSION', '3'
        )
        self.useFixture(ver_fixture)
        auth_url = os.environ.get('OS_AUTH_URL')
        if auth_url:
            auth_url_fixture = fixtures.EnvironmentVariable(
                'OS_AUTH_URL', auth_url.replace('v2.0', 'v3')
            )
            self.useFixture(auth_url_fixture)

    def _create_dummy_user(self, add_clean_up=True):
        username = data_utils.rand_name('TestUser')
        password = data_utils.rand_name('password')
        email = data_utils.rand_name() + '@example.com'
        description = data_utils.rand_name('description')
        raw_output = self.openstack(
            'user create '
            f'--domain {self.domain_name} '
            f'--project {self.project_name} '
            f'--project-domain {self.domain_name} '
            f'--password {password} '
            f'--email {email} '
            f'--description {description} '
            '--enable '
            f'{username}'
        )
        if add_clean_up:
            self.addCleanup(
                self.openstack,
                'user delete {}'.format(
                    self.parse_show_as_object(raw_output)['id']
                ),
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.USER_FIELDS)
        return username

    def _create_dummy_role(self, add_clean_up=True):
        role_name = data_utils.rand_name('TestRole')
        raw_output = self.openstack(f'role create {role_name}')
        role = self.parse_show_as_object(raw_output)
        if add_clean_up:
            self.addCleanup(
                self.openstack, 'role delete {}'.format(role['id'])
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.ROLE_FIELDS)
        self.assertEqual(role_name, role['name'])
        return role_name

    def _create_dummy_implied_role(self, add_clean_up=True):
        role_name = self._create_dummy_role(add_clean_up)
        implied_role_name = self._create_dummy_role(add_clean_up)
        self.openstack(
            'implied role create '
            f'--implied-role {implied_role_name} '
            f'{role_name}'
        )

        return implied_role_name, role_name

    def _create_dummy_group(self, add_clean_up=True):
        group_name = data_utils.rand_name('TestGroup')
        description = data_utils.rand_name('description')
        raw_output = self.openstack(
            'group create '
            f'--domain {self.domain_name} '
            f'--description {description} '
            f'{group_name}'
        )
        if add_clean_up:
            self.addCleanup(
                self.openstack,
                f'group delete --domain {self.domain_name} {group_name}',
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.GROUP_FIELDS)
        return group_name

    def _create_dummy_domain(self, add_clean_up=True):
        domain_name = data_utils.rand_name('TestDomain')
        domain_description = data_utils.rand_name('description')
        self.openstack(
            'domain create '
            f'--description {domain_description} '
            f'--enable {domain_name}'
        )
        if add_clean_up:
            self.addCleanup(self.openstack, f'domain delete {domain_name}')
            self.addCleanup(
                self.openstack, f'domain set --disable {domain_name}'
            )
        return domain_name

    def _create_dummy_project(self, add_clean_up=True):
        project_name = data_utils.rand_name('TestProject')
        project_description = data_utils.rand_name('description')
        self.openstack(
            'project create '
            f'--domain {self.domain_name} '
            f'--description {project_description} '
            f'--enable {project_name}'
        )
        if add_clean_up:
            self.addCleanup(
                self.openstack,
                f'project delete --domain {self.domain_name} {project_name}',
            )
        return project_name

    def _create_dummy_region(self, parent_region=None, add_clean_up=True):
        region_id = data_utils.rand_name('TestRegion')
        description = data_utils.rand_name('description')
        parent_region_arg = ''
        if parent_region is not None:
            parent_region_arg = f'--parent-region {parent_region}'
        raw_output = self.openstack(
            'region create '
            f'{parent_region_arg} '
            f'--description {description} '
            f'{region_id}'
        )
        if add_clean_up:
            self.addCleanup(self.openstack, f'region delete {region_id}')
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.REGION_FIELDS)
        return region_id

    def _create_dummy_service(self, add_clean_up=True):
        service_name = data_utils.rand_name('TestService')
        description = data_utils.rand_name('description')
        type_name = data_utils.rand_name('TestType')
        raw_output = self.openstack(
            'service create '
            f'--name {service_name} '
            f'--description {description} '
            '--enable '
            f'{type_name}'
        )
        if add_clean_up:
            service = self.parse_show_as_object(raw_output)
            self.addCleanup(
                self.openstack, 'service delete {}'.format(service['id'])
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.SERVICE_FIELDS)
        return service_name

    def _create_dummy_endpoint(self, interface='public', add_clean_up=True):
        region_id = self._create_dummy_region()
        service_name = self._create_dummy_service()
        endpoint_url = data_utils.rand_url()
        raw_output = self.openstack(
            'endpoint create '
            f'--region {region_id} '
            '--enable '
            f'{service_name} '
            f'{interface} '
            f'{endpoint_url}'
        )
        endpoint = self.parse_show_as_object(raw_output)
        if add_clean_up:
            self.addCleanup(
                self.openstack, 'endpoint delete {}'.format(endpoint['id'])
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.ENDPOINT_FIELDS)
        return endpoint['id']

    def _create_dummy_idp(self, add_clean_up=True):
        identity_provider = data_utils.rand_name('IdentityProvider')
        description = data_utils.rand_name('description')
        raw_output = self.openstack(
            'identity provider create '
            f' {identity_provider} '
            f'--description {description} '
            '--enable '
        )
        if add_clean_up:
            self.addCleanup(
                self.openstack,
                f'identity provider delete {identity_provider}',
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.IDENTITY_PROVIDER_FIELDS)
        return identity_provider

    def _create_dummy_sp(self, add_clean_up=True):
        service_provider = data_utils.rand_name('ServiceProvider')
        description = data_utils.rand_name('description')
        raw_output = self.openstack(
            'service provider create '
            f' {service_provider} '
            f'--description {description} '
            '--auth-url https://sp.example.com:35357 '
            '--service-provider-url https://sp.example.com:5000 '
            '--enable '
        )
        if add_clean_up:
            self.addCleanup(
                self.openstack, f'service provider delete {service_provider}'
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.SERVICE_PROVIDER_FIELDS)
        return service_provider

    def _create_dummy_registered_limit(self, add_clean_up=True):
        service_name = self._create_dummy_service()
        resource_name = data_utils.rand_name('resource_name')
        params = {
            'service_name': service_name,
            'default_limit': 10,
            'resource_name': resource_name,
        }
        raw_output = self.openstack(
            'registered limit create'
            ' --service {service_name}'
            ' --default-limit {default_limit}'
            ' {resource_name}'.format(**params),
            cloud=SYSTEM_CLOUD,
        )
        items = self.parse_show(raw_output)
        registered_limit_id = self._extract_value_from_items('id', items)

        if add_clean_up:
            self.addCleanup(
                self.openstack,
                f'registered limit delete {registered_limit_id}',
                cloud=SYSTEM_CLOUD,
            )

        self.assert_show_fields(items, self.REGISTERED_LIMIT_FIELDS)
        return registered_limit_id

    def _extract_value_from_items(self, key, items):
        for d in items:
            for k, v in d.items():
                if k == key:
                    return v

    def _create_dummy_limit(self, add_clean_up=True):
        registered_limit_id = self._create_dummy_registered_limit()

        raw_output = self.openstack(
            f'registered limit show {registered_limit_id}',
            cloud=SYSTEM_CLOUD,
        )
        items = self.parse_show(raw_output)
        resource_name = self._extract_value_from_items('resource_name', items)
        service_id = self._extract_value_from_items('service_id', items)
        resource_limit = 15

        project_name = self._create_dummy_project()
        raw_output = self.openstack(f'project show {project_name}')
        items = self.parse_show(raw_output)
        project_id = self._extract_value_from_items('id', items)

        params = {
            'project_id': project_id,
            'service_id': service_id,
            'resource_name': resource_name,
            'resource_limit': resource_limit,
        }

        raw_output = self.openstack(
            'limit create'
            ' --project {project_id}'
            ' --service {service_id}'
            ' --resource-limit {resource_limit}'
            ' {resource_name}'.format(**params),
            cloud=SYSTEM_CLOUD,
        )
        items = self.parse_show(raw_output)
        limit_id = self._extract_value_from_items('id', items)

        if add_clean_up:
            self.addCleanup(
                self.openstack,
                f'limit delete {limit_id}',
                cloud=SYSTEM_CLOUD,
            )

        self.assert_show_fields(items, self.LIMIT_FIELDS)
        return limit_id



#   Copyright 2014 eBay Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstackclient.identity.v3 import token
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestTokenIssue(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        self.cmd = token.IssueToken(self.app, None)

    def test_token_issue_with_project_id(self):
        auth_ref = identity_fakes.fake_auth_ref(
            identity_fakes.TOKEN_WITH_PROJECT_ID,
        )
        self.app.client_manager.auth_ref = auth_ref

        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        collist = ('expires', 'id', 'project_id', 'user_id')
        self.assertEqual(collist, columns)
        datalist = (
            identity_fakes.token_expires,
            identity_fakes.token_id,
            identity_fakes.project_id,
            identity_fakes.user_id,
        )
        self.assertEqual(datalist, data)

    def test_token_issue_with_domain_id(self):
        auth_ref = identity_fakes.fake_auth_ref(
            identity_fakes.TOKEN_WITH_DOMAIN_ID,
        )
        self.app.client_manager.auth_ref = auth_ref

        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        collist = ('domain_id', 'expires', 'id', 'user_id')
        self.assertEqual(collist, columns)
        datalist = (
            identity_fakes.domain_id,
            identity_fakes.token_expires,
            identity_fakes.token_id,
            identity_fakes.user_id,
        )
        self.assertEqual(datalist, data)

    def test_token_issue_with_unscoped(self):
        auth_ref = identity_fakes.fake_auth_ref(
            identity_fakes.UNSCOPED_TOKEN,
        )
        self.app.client_manager.auth_ref = auth_ref

        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        collist = (
            'expires',
            'id',
            'user_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            identity_fakes.token_expires,
            identity_fakes.token_id,
            identity_fakes.user_id,
        )
        self.assertEqual(datalist, data)


class TestTokenRevoke(identity_fakes.TestIdentityv3):
    TOKEN = 'fob'

    def setUp(self):
        super().setUp()
        self.tokens_mock = self.identity_client.tokens
        self.tokens_mock.reset_mock()
        self.tokens_mock.revoke_token.return_value = True
        self.cmd = token.RevokeToken(self.app, None)

    def test_token_revoke(self):
        arglist = [self.TOKEN]
        verifylist = [('token', self.TOKEN)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.tokens_mock.revoke_token.assert_called_with(self.TOKEN)
        self.assertIsNone(result)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Network action implementations"""

from cliff import columns as cliff_columns
from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils
from osc_lib.utils import tags as _tag

from openstackclient.api import compute_v2
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common
from openstackclient.network import common


class AdminStateColumn(cliff_columns.FormattableColumn[bool]):
    def human_readable(self):
        return 'UP' if self._value else 'DOWN'


class RouterExternalColumn(cliff_columns.FormattableColumn[bool]):
    def human_readable(self):
        return 'External' if self._value else 'Internal'


_formatters = {
    'subnets': format_columns.ListColumn,
    'subnet_ids': format_columns.ListColumn,
    'admin_state_up': AdminStateColumn,
    'is_admin_state_up': AdminStateColumn,
    'router:external': RouterExternalColumn,
    'is_router_external': RouterExternalColumn,
    'availability_zones': format_columns.ListColumn,
    'availability_zone_hints': format_columns.ListColumn,
    'tags': format_columns.ListColumn,
}


def _get_columns_network(item):
    column_map = {
        'subnet_ids': 'subnets',
        'is_admin_state_up': 'admin_state_up',
        'is_router_external': 'router:external',
        'is_port_security_enabled': 'port_security_enabled',
        'provider_network_type': 'provider:network_type',
        'provider_physical_network': 'provider:physical_network',
        'provider_segmentation_id': 'provider:segmentation_id',
        'is_shared': 'shared',
        'ipv4_address_scope_id': 'ipv4_address_scope',
        'ipv6_address_scope_id': 'ipv6_address_scope',
        'tags': 'tags',
    }
    hidden_columns = ['location', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


def _get_columns_compute(item):
    return utils.get_osc_show_columns_for_sdk_resource(item, {})


def _get_attrs_network(client_manager, parsed_args):
    attrs = {}
    if parsed_args.name is not None:
        attrs['name'] = parsed_args.name
    if parsed_args.enable:
        attrs['admin_state_up'] = True
    if parsed_args.disable:
        attrs['admin_state_up'] = False
    if parsed_args.share:
        attrs['shared'] = True
    if parsed_args.no_share:
        attrs['shared'] = False
    if parsed_args.enable_port_security:
        attrs['port_security_enabled'] = True
    if parsed_args.disable_port_security:
        attrs['port_security_enabled'] = False

    # "network set" command doesn't support setting project.
    if 'project' in parsed_args and parsed_args.project is not None:
        identity_client = client_manager.identity
        project_id = identity_common.find_project(
            identity_client,
            parsed_args.project,
            parsed_args.project_domain,
        ).id
        attrs['project_id'] = project_id

    # "network set" command doesn't support setting availability zone hints.
    if (
        'availability_zone_hints' in parsed_args
        and parsed_args.availability_zone_hints is not None
    ):
        attrs['availability_zone_hints'] = parsed_args.availability_zone_hints

    # set description
    if parsed_args.description is not None:
        attrs['description'] = parsed_args.description

    # set mtu
    if parsed_args.mtu:
        attrs['mtu'] = parsed_args.mtu

    # update_external_network_options
    if parsed_args.internal:
        attrs['router:external'] = False
    if parsed_args.external:
        attrs['router:external'] = True
    if parsed_args.no_default:
        attrs['is_default'] = False
    if parsed_args.default:
        attrs['is_default'] = True
    # Update Provider network options
    if parsed_args.provider_network_type:
        attrs['provider:network_type'] = parsed_args.provider_network_type
    if parsed_args.physical_network:
        attrs['provider:physical_network'] = parsed_args.physical_network
    if parsed_args.segmentation_id:
        attrs['provider:segmentation_id'] = parsed_args.segmentation_id
    if parsed_args.qos_policy is not None:
        network_client = client_manager.network
        _qos_policy = network_client.find_qos_policy(
            parsed_args.qos_policy, ignore_missing=False
        )
        attrs['qos_policy_id'] = _qos_policy.id
    if 'no_qos_policy' in parsed_args and parsed_args.no_qos_policy:
        attrs['qos_policy_id'] = None
    # Update DNS network options
    if parsed_args.dns_domain is not None:
        attrs['dns_domain'] = parsed_args.dns_domain
    return attrs


def _get_attrs_compute(client_manager, parsed_args):
    attrs = {}
    if parsed_args.name is not None:
        attrs['name'] = parsed_args.name
    if parsed_args.share:
        attrs['share_subnet'] = True
    if parsed_args.no_share:
        attrs['share_subnet'] = False
    if parsed_args.subnet is not None:
        attrs['subnet'] = parsed_args.subnet
    return attrs


def _add_additional_network_options(parser):
    # Add additional network options

    parser.add_argument(
        '--provider-network-type',
        metavar='<provider-network-type>',
        help=_(
            "The physical mechanism by which the virtual network "
            "is implemented. For example: "
            "flat, geneve, gre, local, vlan or vxlan."
        ),
    )
    parser.add_argument(
        '--provider-physical-network',
        metavar='<provider-physical-network>',
        dest='physical_network',
        help=_(
            "Name of the physical network over which the virtual "
            "network is implemented"
        ),
    )
    parser.add_argument(
        '--provider-segment',
        metavar='<provider-segment>',
        dest='segmentation_id',
        help=_(
            "VLAN ID for VLAN networks or Tunnel ID for "
            "GENEVE/GRE/VXLAN networks"
        ),
    )
    parser.add_argument(
        '--dns-domain',
        metavar='<dns-domain>',
        dest='dns_domain',
        help=_(
            "Set DNS domain for this network "
            "(requires DNS integration extension)"
        ),
    )


# TODO(sindhu): Use the SDK resource mapped attribute names once the
# OSC minimum requirements include SDK 1.0.
class CreateNetwork(
    common.NetworkAndComputeShowOne, common.NeutronCommandWithExtraArgs
):
    _description = _("Create new network")

    def update_parser_common(self, parser):
        parser.add_argument(
            'name', metavar='<name>', help=_("New network name")
        )
        share_group = parser.add_mutually_exclusive_group()
        share_group.add_argument(
            '--share',
            action='store_true',
            default=None,
            help=_("Share the network between projects"),
        )
        share_group.add_argument(
            '--no-share',
            action='store_true',
            help=_("Do not share the network between projects"),
        )
        return parser

    def update_parser_network(self, parser):
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable',
            action='store_true',
            default=True,
            help=self.enhance_help_neutron(_("Enable network (default)")),
        )
        admin_group.add_argument(
            '--disable',
            action='store_true',
            help=self.enhance_help_neutron(_("Disable network")),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=self.enhance_help_neutron(_("Owner's project (name or ID)")),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=self.enhance_help_neutron(_("Set network description")),
        )
        parser.add_argument(
            '--mtu',
            metavar='<mtu>',
            help=self.enhance_help_neutron(_("Set network mtu")),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--availability-zone-hint',
            action='append',
            dest='availability_zone_hints',
            metavar='<availability-zone>',
            help=self.enhance_help_neutron(
                _(
                    "Availability Zone in which to create this network "
                    "(Network Availability Zone extension required, "
                    "repeat option to set multiple availability zones)"
                )
            ),
        )
        port_security_group = parser.add_mutually_exclusive_group()
        port_security_group.add_argument(
            '--enable-port-security',
            action='store_true',
            help=self.enhance_help_neutron(
                _(
                    "Enable port security by default for ports created on "
                    "this network (default)"
                )
            ),
        )
        port_security_group.add_argument(
            '--disable-port-security',
            action='store_true',
            help=self.enhance_help_neutron(
                _(
                    "Disable port security by default for ports created on "
                    "this network"
                )
            ),
        )
        external_router_grp = parser.add_mutually_exclusive_group()
        external_router_grp.add_argument(
            '--external',
            action='store_true',
            help=self.enhance_help_neutron(
                _(
                    "The network has an external routing facility that is not "
                    "managed by Neutron and can be used. For example: "
                    "openstack router set --external-gateway NETWORK "
                    "(external-net extension required)"
                )
            ),
        )
        external_router_grp.add_argument(
            '--internal',
            action='store_true',
            help=self.enhance_help_neutron(
                _("Opposite of '--external' (default)")
            ),
        )
        default_router_grp = parser.add_mutually_exclusive_group()
        default_router_grp.add_argument(
            '--default',
            action='store_true',
            help=self.enhance_help_neutron(
                _(
                    "Specify if this network should be used as the default "
                    "external network"
                )
            ),
        )
        default_router_grp.add_argument(
            '--no-default',
            action='store_true',
            help=self.enhance_help_neutron(
                _(
                    "Do not use the network as the default external network "
                    "(default)"
                )
            ),
        )
        parser.add_argument(
            '--qos-policy',
            metavar='<qos-policy>',
            help=self.enhance_help_neutron(
                _("QoS policy to attach to this network (name or ID)")
            ),
        )
        vlan_transparent_grp = parser.add_mutually_exclusive_group()
        vlan_transparent_grp.add_argument(
            '--transparent-vlan',
            action='store_true',
            help=self.enhance_help_neutron(
                _("Make the network VLAN transparent")
            ),
        )
        vlan_transparent_grp.add_argument(
            '--no-transparent-vlan',
            action='store_true',
            help=self.enhance_help_neutron(
                _("Do not make the network VLAN transparent")
            ),
        )

        vlan_qinq_grp = parser.add_mutually_exclusive_group()
        vlan_qinq_grp.add_argument(
            '--qinq-vlan',
            action='store_true',
            help=self.enhance_help_neutron(
                _("Enable VLAN QinQ (S-Tag ethtype 0x8a88) for the network")
            ),
        )
        vlan_qinq_grp.add_argument(
            '--no-qinq-vlan',
            action='store_true',
            help=self.enhance_help_neutron(
                _("Disable VLAN QinQ (S-Tag ethtype 0x8a88) for the network")
            ),
        )

        _add_additional_network_options(parser)
        _tag.add_tag_option_to_parser_for_create(
            parser, _('network'), enhance_help=self.enhance_help_neutron
        )
        return parser

    def update_parser_compute(self, parser):
        parser.add_argument(
            '--subnet',
            metavar='<subnet>',
            required=True,
            help=self.enhance_help_nova_network(
                _("IPv4 subnet for fixed IPs (in CIDR notation)")
            ),
        )
        return parser

    def take_action_network(self, client, parsed_args):
        attrs = _get_attrs_network(self.app.client_manager, parsed_args)
        if parsed_args.transparent_vlan:
            attrs['vlan_transparent'] = True
        if parsed_args.no_transparent_vlan:
            attrs['vlan_transparent'] = False

        if parsed_args.qinq_vlan:
            attrs['vlan_qinq'] = True
        if parsed_args.no_qinq_vlan:
            attrs['vlan_qinq'] = False

        if attrs.get('vlan_transparent') and attrs.get('vlan_qinq'):
            msg = _(
                "--transparent-vlan and --qinq-vlan can not be both enabled "
                "for the network."
            )
            raise exceptions.CommandError(msg)

        if (
            parsed_args.segmentation_id
            and not parsed_args.provider_network_type
        ):
            msg = _(
                "--provider-segment requires --provider-network-type "
                "to be specified."
            )
            raise exceptions.CommandError(msg)

        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )
        with common.check_missing_extension_if_error(
            self.app.client_manager.network, attrs
        ):
            obj = client.create_network(**attrs)

        # tags cannot be set when created, so tags need to be set later.
        _tag.update_tags_for_set(client, obj, parsed_args)
        display_columns, columns = _get_columns_network(obj)
        data = utils.get_item_properties(obj, columns, formatters=_formatters)
        return (display_columns, data)

    def take_action_compute(self, client, parsed_args):
        attrs = _get_attrs_compute(self.app.client_manager, parsed_args)
        obj = compute_v2.create_network(client, **attrs)
        display_columns, columns = _get_columns_compute(obj)
        data = utils.get_dict_properties(obj, columns)
        return (display_columns, data)


class DeleteNetwork(common.NetworkAndComputeDelete):
    _description = _("Delete network(s)")

    # Used by base class to find resources in parsed_args.
    resource = 'network'
    r = None

    def update_parser_common(self, parser):
        parser.add_argument(
            'network',
            metavar="<network>",
            nargs="+",
            help=_("Network(s) to delete (name or ID)"),
        )

        return parser

    def take_action_network(self, client, parsed_args):
        obj = client.find_network(self.r, ignore_missing=False)
        client.delete_network(obj)

    def take_action_compute(self, client, parsed_args):
        network = compute_v2.find_network(client, self.r)
        compute_v2.delete_network(client, network['id'])


# TODO(sindhu): Use the SDK resource mapped attribute names once the
# OSC minimum requirements include SDK 1.0.
class ListNetwork(common.NetworkAndComputeLister):
    _description = _("List networks")

    def update_parser_network(self, parser):
        router_ext_group = parser.add_mutually_exclusive_group()
        router_ext_group.add_argument(
            '--external',
            action='store_true',
            help=self.enhance_help_neutron(_("List only external networks")),
        )
        router_ext_group.add_argument(
            '--internal',
            action='store_true',
            help=self.enhance_help_neutron(_("List only internal networks")),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            help=self.enhance_help_neutron(
                _("List additional fields in output")
            ),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=self.enhance_help_neutron(
                _("List only networks with the specified name")
            ),
        )
        admin_state_group = parser.add_mutually_exclusive_group()
        admin_state_group.add_argument(
            '--enable',
            action='store_true',
            help=self.enhance_help_neutron(_("List only enabled networks")),
        )
        admin_state_group.add_argument(
            '--disable',
            action='store_true',
            help=self.enhance_help_neutron(_("List only disabled networks")),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                "List only networks with the specified project (name or ID)"
            ),
        )
        identity_common.add_project_domain_option_to_parser(
            parser, enhance_help=self.enhance_help_neutron
        )
        shared_group = parser.add_mutually_exclusive_group()
        shared_group.add_argument(
            '--share',
            action='store_true',
            help=self.enhance_help_neutron(
                _("List only networks shared between projects")
            ),
        )
        shared_group.add_argument(
            '--no-share',
            action='store_true',
            help=self.enhance_help_neutron(
                _("List only networks not shared between projects")
            ),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            choices=['ACTIVE', 'BUILD', 'DOWN', 'ERROR'],
            help=self.enhance_help_neutron(
                _(
                    "List only networks with the specified status "
                    "('ACTIVE', 'BUILD', 'DOWN', 'ERROR')"
                )
            ),
        )
        parser.add_argument(
            '--provider-network-type',
            metavar='<provider-network-type>',
            choices=['flat', 'geneve', 'gre', 'local', 'vlan', 'vxlan'],
            help=self.enhance_help_neutron(
                _(
                    "List only networks with the specified physical "
                    "mechanisms. "
                    "The supported options are: flat, geneve, gre, local, "
                    "vlan and vxlan."
                )
            ),
        )
        parser.add_argument(
            '--provider-physical-network',
            metavar='<provider-physical-network>',
            dest='physical_network',
            help=self.enhance_help_neutron(
                _(
                    "List only networks with the specified physical network "
                    "name"
                )
            ),
        )
        parser.add_argument(
            '--provider-segment',
            metavar='<provider-segment>',
            dest='segmentation_id',
            help=self.enhance_help_neutron(
                _(
                    "List only networks with the specified provider segment "
                    "ID (VLAN ID for VLAN networks or "
                    "Tunnel ID for GENEVE/GRE/VXLAN networks)"
                )
            ),
        )
        parser.add_argument(
            '--agent',
            metavar='<agent-id>',
            dest='agent_id',
            help=self.enhance_help_neutron(
                _('List only networks hosted the specified agent (ID only)')
            ),
        )
        _tag.add_tag_filtering_option_to_parser(
            parser, _('networks'), enhance_help=self.enhance_help_neutron
        )
        return parser

    def take_action_network(self, client, parsed_args):
        identity_client = self.app.client_manager.identity
        if parsed_args.long:
            columns: tuple[str, ...] = (
                'id',
                'name',
                'status',
                'project_id',
                'is_admin_state_up',
                'is_shared',
                'subnet_ids',
                'provider_network_type',
                'is_router_external',
                'availability_zones',
                'tags',
            )
            column_headers: tuple[str, ...] = (
                'ID',
                'Name',
                'Status',
                'Project',
                'State',
                'Shared',
                'Subnets',
                'Network Type',
                'Router Type',
                'Availability Zones',
                'Tags',
            )
        elif parsed_args.agent_id:
            columns = ('id', 'name', 'subnet_ids')
            column_headers = (
                'ID',
                'Name',
                'Subnets',
            )
            client = self.app.client_manager.network
            dhcp_agent = client.get_agent(parsed_args.agent_id)
            data = client.dhcp_agent_hosting_networks(dhcp_agent)

            return (
                column_headers,
                (
                    utils.get_item_properties(
                        s,
                        columns,
                        formatters=_formatters,
                    )
                    for s in data
                ),
            )
        else:
            columns = ('id', 'name', 'subnet_ids')
            column_headers = (
                'ID',
                'Name',
                'Subnets',
            )

        args = {}

        if parsed_args.external:
            args['router:external'] = True
            args['is_router_external'] = True
        elif parsed_args.internal:
            args['router:external'] = False
            args['is_router_external'] = False

        if parsed_args.name is not None:
            args['name'] = parsed_args.name

        if parsed_args.enable:
            args['admin_state_up'] = True
            args['is_admin_state_up'] = True
        elif parsed_args.disable:
            args['admin_state_up'] = False
            args['is_admin_state_up'] = False

        if parsed_args.project:
            project = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            args['project_id'] = project.id

        if parsed_args.share:
            args['shared'] = True
            args['is_shared'] = True
        elif parsed_args.no_share:
            args['shared'] = False
            args['is_shared'] = False

        if parsed_args.status:
            args['status'] = parsed_args.status

        if parsed_args.provider_network_type:
            args['provider:network_type'] = parsed_args.provider_network_type
            args['provider_network_type'] = parsed_args.provider_network_type
        if parsed_args.physical_network:
            args['provider:physical_network'] = parsed_args.physical_network
            args['provider_physical_network'] = parsed_args.physical_network
        if parsed_args.segmentation_id:
            args['provider:segmentation_id'] = parsed_args.segmentation_id
            args['provider_segmentation_id'] = parsed_args.segmentation_id

        _tag.get_tag_filtering_args(parsed_args, args)

        data = client.networks(**args)

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters=_formatters,
                )
                for s in data
            ),
        )

    def take_action_compute(self, client, parsed_args):
        columns = (
            'id',
            'label',
            'cidr',
        )
        column_headers = (
            'ID',
            'Name',
            'Subnet',
        )

        data = compute_v2.list_networks(client)

        return (
            column_headers,
            (
                utils.get_dict_properties(
                    s,
                    columns,
                    formatters=_formatters,
                )
                for s in data
            ),
        )


# TODO(sindhu): Use the SDK resource mapped attribute names once the
# OSC minimum requirements include SDK 1.0.
class SetNetwork(common.NeutronCommandWithExtraArgs):
    _description = _("Set network properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar="<network>",
            help=_("Network to modify (name or ID)"),
        )
        parser.add_argument(
            '--name', metavar='<name>', help=_("Set network name")
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable',
            action='store_true',
            default=None,
            help=_("Enable network"),
        )
        admin_group.add_argument(
            '--disable', action='store_true', help=_("Disable network")
        )
        share_group = parser.add_mutually_exclusive_group()
        share_group.add_argument(
            '--share',
            action='store_true',
            default=None,
            help=_("Share the network between projects"),
        )
        share_group.add_argument(
            '--no-share',
            action='store_true',
            help=_("Do not share the network between projects"),
        )
        parser.add_argument(
            '--description',
            metavar="<description>",
            help=_("Set network description"),
        )
        parser.add_argument(
            '--mtu', metavar="<mtu>", help=_("Set network mtu")
        )
        port_security_group = parser.add_mutually_exclusive_group()
        port_security_group.add_argument(
            '--enable-port-security',
            action='store_true',
            help=_(
                "Enable port security by default for ports created on "
                "this network"
            ),
        )
        port_security_group.add_argument(
            '--disable-port-security',
            action='store_true',
            help=_(
                "Disable port security by default for ports created on "
                "this network"
            ),
        )
        external_router_grp = parser.add_mutually_exclusive_group()
        external_router_grp.add_argument(
            '--external',
            action='store_true',
            help=_(
                "The network has an external routing facility that is not "
                "managed by Neutron and can be used. For example: "
                "openstack router set --external-gateway NETWORK "
                "(external-net extension required)."
            ),
        )
        external_router_grp.add_argument(
            '--internal',
            action='store_true',
            help=_("Opposite of '--external'"),
        )
        default_router_grp = parser.add_mutually_exclusive_group()
        default_router_grp.add_argument(
            '--default',
            action='store_true',
            help=_("Set the network as the default external network"),
        )
        default_router_grp.add_argument(
            '--no-default',
            action='store_true',
            help=_("Do not use the network as the default external network"),
        )
        qos_group = parser.add_mutually_exclusive_group()
        qos_group.add_argument(
            '--qos-policy',
            metavar='<qos-policy>',
            help=_("QoS policy to attach to this network (name or ID)"),
        )
        qos_group.add_argument(
            '--no-qos-policy',
            action='store_true',
            help=_("Remove the QoS policy attached to this network"),
        )
        _tag.add_tag_option_to_parser_for_set(parser, _('network'))
        _add_additional_network_options(parser)
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        obj = client.find_network(parsed_args.network, ignore_missing=False)

        attrs = _get_attrs_network(self.app.client_manager, parsed_args)
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )
        if attrs:
            with common.check_missing_extension_if_error(
                self.app.client_manager.network, attrs
            ):
                client.update_network(obj, **attrs)

        # tags is a subresource and it needs to be updated separately.
        _tag.update_tags_for_set(client, obj, parsed_args)


class ShowNetwork(common.NetworkAndComputeShowOne):
    _description = _("Show network details")

    def update_parser_common(self, parser):
        parser.add_argument(
            'network',
            metavar="<network>",
            help=_("Network to display (name or ID)"),
        )
        return parser

    def take_action_network(self, client, parsed_args):
        obj = client.find_network(parsed_args.network, ignore_missing=False)
        display_columns, columns = _get_columns_network(obj)
        data = utils.get_item_properties(obj, columns, formatters=_formatters)
        return (display_columns, data)

    def take_action_compute(self, client, parsed_args):
        obj = compute_v2.find_network(client, parsed_args.network)
        display_columns, columns = _get_columns_compute(obj)
        data = utils.get_dict_properties(obj, columns)
        return (display_columns, data)


class UnsetNetwork(common.NeutronUnsetCommandWithExtraArgs):
    _description = _("Unset network properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar="<network>",
            help=_("Network to modify (name or ID)"),
        )
        _tag.add_tag_option_to_parser_for_unset(parser, _('network'))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        obj = client.find_network(parsed_args.network, ignore_missing=False)

        attrs = self._parse_extra_properties(parsed_args.extra_properties)
        if attrs:
            client.update_network(obj, **attrs)

        # tags is a subresource and it needs to be updated separately.
        _tag.update_tags_for_unset(client, obj, parsed_args)
